import unittest

import numpy as np
import pandas as pd
import shutil
from pathlib import Path

from aethos.core import Data


class Test_TestBase(unittest.TestCase):
    @classmethod
    def tearDownClass(cls):
        report_path = str(Path.home()) + "/.aethos/reports/"

        if Path(report_path).exists():
            shutil.rmtree(report_path)

    def test_compare_dist_predict(self):

        data = np.random.randint(0, 2, size=(1000, 3))
        data = pd.DataFrame(data=data, columns=["col1", "col2", "col3"])

        df = Data(data, target_field="col3", report_name="test")
        df.predict_data_sample()

        self.assertTrue(True)

    def test_compare_dist_ks(self):

        data = np.random.randint(0, 2, size=(1000, 3))
        data = pd.DataFrame(data=data, columns=["col1", "col2", "col3"])

        df = Data(data, target_field="col3", report_name="test")
        df["col4"] = np.random.normal(1, 2, size=(1, 800))[0]
        df["col4"] = np.random.normal(10, 20, size=(1, 200))[0]
        df.ks_feature_distribution()

        self.assertTrue(True)

    def test_most_common_list(self):

        data = pd.Series([["hi", "aethos"], ["hi", "py-automl"], [], ["hi"]])
        data = pd.DataFrame(data, columns=["col1"])

        df = Data(data, split=False)
        df.most_common("col1", plot=True)

        self.assertTrue(True)

    def test_most_common_str(self):

        data = pd.Series(
            [
                "hi aethos",
                "aethos is awesome",
                "hi",
                "py-automl is the old name",
                "hi everyone",
            ]
        )
        data = pd.DataFrame(data, columns=["col1"])

        df = Data(data, split=False)
        df.most_common("col1")

        self.assertTrue(True)

    def test_most_common_int(self):

        data = pd.Series([1, 1, 2, 4, 2, 5])
        data = pd.DataFrame(data, columns=["col1"])

        df = Data(data, test_split_percentage=0.5)
        df.most_common("col1", plot=True, use_test=True)

        self.assertTrue(True)

    def test_2s_pairedttest(self):

        data1 = np.random.normal(0, 1, size=50)
        data2 = np.random.normal(2, 1, size=50)

        data = pd.DataFrame({"d1": data1, "d2": data2})

        df = Data(data)

        df.paired_ttest("d1", "d2")

        self.assertTrue(True)

    def test_2s_indttest(self):

        data1 = np.random.normal(0, 1, size=50)
        data2 = np.random.normal(2, 1, size=50)

        data = pd.DataFrame({"d1": data1, "d2": data2})

        df = Data(data)

        df.ind_ttest("d1", "d2")

        self.assertTrue(True)

    def test_1s_indttest(self):

        data1 = np.random.normal(0, 1, size=50)
        data2 = np.random.normal(2, 1, size=50)

        data = pd.DataFrame({"d1": data1, "d2": data2})

        df = Data(data)

        df.onesample_ttest("d1", 1.0)

        self.assertTrue(True)

    def test_anova_asserterror(self):

        data1 = np.random.normal(0, 1, size=50)
        data2 = np.random.normal(2, 1, size=50)

        data = pd.DataFrame({"d1": data1, "d2": data2})

        df = Data(data)

        self.assertRaises(AssertionError, df.anova, "d1")

    def test_anova_numvar(self):

        int_missing_data = [
            [1, 8, "blue", 1],
            [2, 8, "blue", 0],
            [3, 8, "green", 0],
            [4, 8, "green", 1],
            [5, 8, "yellow", 1],
            [6, 8, "yellow", 0],
            [1, 8, "blue", 1],
            [2, 8, "blue", 0],
            [3, 8, "green", 0],
            [4, 8, "green", 1],
            [5, 8, "yellow", 1],
            [6, 8, "yellow", 0],
            [1, 8, "blue", 1],
            [2, 8, "blue", 0],
            [3, 8, "green", 0],
            [4, 8, "green", 1],
            [5, 8, "yellow", 1],
            [6, 8, "yellow", 0],
            [1, 8, "blue", 1],
            [2, 8, "blue", 0],
            [3, 8, "green", 0],
            [4, 8, "green", 1],
            [5, 8, "yellow", 1],
            [6, 8, "yellow", 0],
            [1, 8, "blue", 1],
            [2, 8, "blue", 0],
            [3, 8, "green", 0],
            [4, 8, "green", 1],
            [5, 8, "yellow", 1],
            [6, 8, "yellow", 0],
            [1, 8, "blue", 1],
            [2, 8, "blue", 0],
            [3, 8, "green", 0],
            [4, 8, "green", 1],
            [5, 8, "yellow", 1],
            [6, 8, "yellow", 0],
            [1, 8, "blue", 1],
            [2, 8, "blue", 0],
            [3, 8, "green", 0],
            [4, 8, "green", 1],
            [5, 8, "yellow", 1],
            [6, 8, "yellow", 0],
            [1, 8, "blue", 1],
            [2, 8, "blue", 0],
            [3, 8, "green", 0],
            [4, 8, "green", 1],
            [5, 8, "yellow", 1],
            [6, 8, "yellow", 0],
        ]

        columns = ["col1", "col2", "col3", "col4"]
        data = pd.DataFrame(int_missing_data, columns=columns)

        feat = Data(
            x_train=data,
            target_field="col4",
            test_split_percentage=0.5,
            report_name="test",
        )

        feat.anova("col4", num_variables=["col1", "col2"], verbose=True)

        self.assertTrue(True)

    def test_anova_catvar(self):

        int_missing_data = [
            [1, 8, "blue", 1],
            [2, 8, "blue", 0],
            [3, 8, "green", 0],
            [4, 8, "green", 1],
            [5, 8, "yellow", 1],
            [6, 8, "yellow", 0],
            [1, 8, "blue", 1],
            [2, 8, "blue", 0],
            [3, 8, "green", 0],
            [4, 8, "green", 1],
            [5, 8, "yellow", 1],
            [6, 8, "yellow", 0],
            [1, 8, "blue", 1],
            [2, 8, "blue", 0],
            [3, 8, "green", 0],
            [4, 8, "green", 1],
            [5, 8, "yellow", 1],
            [6, 8, "yellow", 0],
            [1, 8, "blue", 1],
            [2, 8, "blue", 0],
            [3, 8, "green", 0],
            [4, 8, "green", 1],
            [5, 8, "yellow", 1],
            [6, 8, "yellow", 0],
            [1, 8, "blue", 1],
            [2, 8, "blue", 0],
            [3, 8, "green", 0],
            [4, 8, "green", 1],
            [5, 8, "yellow", 1],
            [6, 8, "yellow", 0],
            [1, 8, "blue", 1],
            [2, 8, "blue", 0],
            [3, 8, "green", 0],
            [4, 8, "green", 1],
            [5, 8, "yellow", 1],
            [6, 8, "yellow", 0],
            [1, 8, "blue", 1],
            [2, 8, "blue", 0],
            [3, 8, "green", 0],
            [4, 8, "green", 1],
            [5, 8, "yellow", 1],
            [6, 8, "yellow", 0],
            [1, 8, "blue", 1],
            [2, 8, "blue", 0],
            [3, 8, "green", 0],
            [4, 8, "green", 1],
            [5, 8, "yellow", 1],
            [6, 8, "yellow", 0],
        ]

        columns = ["col1", "col2", "col3", "col4"]
        data = pd.DataFrame(int_missing_data, columns=columns)

        feat = Data(
            x_train=data,
            target_field="col4",
            test_split_percentage=0.5,
            report_name="test",
        )

        feat.anova("col4", cat_variables=["col3"], verbose=True)

        self.assertTrue(True)

    def test_anova(self):

        int_missing_data = [
            [1, 8, "blue", 1],
            [2, 8, "blue", 0],
            [3, 8, "green", 0],
            [4, 8, "green", 1],
            [5, 8, "yellow", 1],
            [6, 8, "yellow", 0],
            [1, 8, "blue", 1],
            [2, 8, "blue", 0],
            [3, 8, "green", 0],
            [4, 8, "green", 1],
            [5, 8, "yellow", 1],
            [6, 8, "yellow", 0],
            [1, 8, "blue", 1],
            [2, 8, "blue", 0],
            [3, 8, "green", 0],
            [4, 8, "green", 1],
            [5, 8, "yellow", 1],
            [6, 8, "yellow", 0],
            [1, 8, "blue", 1],
            [2, 8, "blue", 0],
            [3, 8, "green", 0],
            [4, 8, "green", 1],
            [5, 8, "yellow", 1],
            [6, 8, "yellow", 0],
            [1, 8, "blue", 1],
            [2, 8, "blue", 0],
            [3, 8, "green", 0],
            [4, 8, "green", 1],
            [5, 8, "yellow", 1],
            [6, 8, "yellow", 0],
            [1, 8, "blue", 1],
            [2, 8, "blue", 0],
            [3, 8, "green", 0],
            [4, 8, "green", 1],
            [5, 8, "yellow", 1],
            [6, 8, "yellow", 0],
            [1, 8, "blue", 1],
            [2, 8, "blue", 0],
            [3, 8, "green", 0],
            [4, 8, "green", 1],
            [5, 8, "yellow", 1],
            [6, 8, "yellow", 0],
            [1, 8, "blue", 1],
            [2, 8, "blue", 0],
            [3, 8, "green", 0],
            [4, 8, "green", 1],
            [5, 8, "yellow", 1],
            [6, 8, "yellow", 0],
        ]

        columns = ["col1", "col2", "col3", "col4"]
        data = pd.DataFrame(int_missing_data, columns=columns)

        feat = Data(
            x_train=data,
            target_field="col4",
            test_split_percentage=0.5,
            report_name="test",
        )

        feat.anova(
            "col4", num_variables=["col1", "col2"], cat_variables=["col3"], verbose=True
        )

        self.assertTrue(True)



import string
from functools import partial

import pandas as pd
from nltk import sent_tokenize
from nltk.corpus import stopwords
from nltk.stem import WordNetLemmatizer
from nltk.stem.snowball import PorterStemmer, SnowballStemmer
from nltk.tokenize import RegexpTokenizer, word_tokenize

NLTK_STEMMERS = {"porter": PorterStemmer(), "snowball": SnowballStemmer("english")}

NLTK_LEMMATIZERS = {"wordnet": WordNetLemmatizer()}


def split_sentences(x_train, x_test=None, list_of_cols=[], new_col_name="_sentences"):
    """
    Splits text by its sentences and then saves that list in a new column.

    Parameters
    ----------
    x_train : DataFrame
        Dataset

    x_test : DataFrame
        Testing dataset, by default None

    list_of_cols : list, optional
        Column name(s) of text data that you want to separate into sentences

    new_col_name : str, optional
        New column name to be created when applying this technique, by default `COLUMN_sentences`

    Returns
    -------
    Dataframe, *Dataframe
        Transformed dataframe with the new column

    Returns 2 Dataframes if x_test is provided. 
    """

    for col in list_of_cols:
        if new_col_name.startswith("_"):
            new_col_name = col + new_col_name

        x_train[new_col_name] = pd.Series(map(sent_tokenize, x_train[col]))
        if x_test is not None:
            x_test[new_col_name] = pd.Series(map(sent_tokenize, x_test[col]))

    return x_train, x_test


def nltk_stem(
    x_train, x_test=None, list_of_cols=[], stemmer="porter", new_col_name="_stemmed"
):
    """
    Stems all the text in a column.

    Parameters
    ----------
    x_train : DataFrame
        Dataset

    x_test : DataFrame
        Testing dataset, by default None

    list_of_cols : list, optional
        Column name(s) of text data that you want to stem

    stemmer : str, optional
        Type of NLTK stemmer to use, by default porter

        Current stemming implementations:
            - Porter

        For more information please refer to the NLTK stemming api https://www.nltk.org/api/nltk.stem.html

    new_col_name : str, optional
        New column name to be created when applying this technique, by default `COLUMN_stemmed`

    Returns
    -------
    Dataframe, *Dataframe
        Transformed dataframe with the new column.

    Returns 2 Dataframes if x_test is provided. 
    """

    stem = NLTK_STEMMERS[stemmer]
    # Create partial for speed purposes
    func = partial(_apply_text_method, transformer=stem.stem)

    for col in list_of_cols:
        if new_col_name.startswith("_"):
            new_col_name = col + new_col_name

        x_train[new_col_name] = pd.Series(map(func, x_train[col]))

        if x_test is not None:
            x_test[new_col_name] = pd.Series(map(func, x_test[col]))

    return x_train, x_test


def nltk_word_tokenizer(
    x_train, x_test=None, list_of_cols=[], regexp="", new_col_name="_tokenized"
):
    """
    Splits text into its words. 

    Default is by spaces but if a regex expression is provided, it will use that.

    Parameters
    ----------
    x_train : DataFrame
        Dataset
        
    x_test : DataFrame
        Testing dataset, by default None

    list_of_cols : list, optional
        Column name(s) of text data that you want to stem

    pattern : str, optional
        Regex pattern used to split words.

    new_col_name : str, optional
        New column name to be created when applying this technique, by default `COLUMN_tokenized`

    Returns
    -------
    Dataframe, *Dataframe
        Transformed dataframe with the new column.

    Returns 2 Dataframes if x_test is provided. 
    """

    tokenizer = RegexpTokenizer(regexp)

    for col in list_of_cols:
        if new_col_name.startswith("_"):
            new_col_name = col + new_col_name

        if not regexp:
            x_train[new_col_name] = pd.Series(map(word_tokenize, x_train[col]))

            if x_test is not None:
                x_test[new_col_name] = pd.Series(map(word_tokenize, x_test[col]))
        else:
            x_train[new_col_name] = pd.Series(map(tokenizer.tokenize, x_train[col]))

            if x_test is not None:
                x_test[new_col_name] = pd.Series(map(tokenizer.tokenize, x_test[col]))

    return x_train, x_test


def nltk_remove_stopwords(
    x_train, x_test=None, list_of_cols=[], custom_stopwords=[], new_col_name="_rem_stop"
):
    """
    Removes stopwords following the nltk English stopwords list.

    A list of custom words can be provided as well, usually for domain specific words.

    Parameters
    ----------
    x_train : DataFrame
        Dataset
        
    x_test : DataFrame
        Testing dataset, by default None

    list_of_cols : list, optional
        Column name(s) of text data that you want to stem

    regexp : str, optional
        Regex pattern used to split words.

    new_col_name : str, optional
        New column name to be created when applying this technique, by default `COLUMN_rem_words`

    Returns
    -------
    Dataframe, *Dataframe
        Transformed dataframe with the new column.

    Returns 2 Dataframes if x_test is provided. 
    """

    stop_words = stopwords.words("english")
    stop_words.extend(custom_stopwords)
    stop_list = set(stop_words)

    for col in list_of_cols:
        if new_col_name.startswith("_"):
            new_col_name = col + new_col_name

        x_train[new_col_name] = list(
            map(
                lambda x: " ".join(
                    [word for word in word_tokenize(x.lower()) if word not in stop_list]
                ),
                x_train[col],
            )
        )

        if x_test is not None:
            x_test[new_col_name] = list(
                map(
                    lambda x: " ".join(
                        [
                            word
                            for word in word_tokenize(x.lower())
                            if word not in stop_list
                        ]
                    ),
                    x_test[col],
                )
            )

    return x_train, x_test


def remove_punctuation(
    x_train,
    x_test=None,
    list_of_cols=[],
    regexp="",
    exceptions=[],
    new_col_name="_rem_punct",
):
    """
    Removes punctuation from a string.

    Defaults to removing all punctuation, but if regex of punctuation is provided, it will remove them.

    Parameters
    ----------
    x_train : DataFrame
        Dataset

    x_test : DataFrame
        Testing dataset, by default None

    list_of_cols : list, optional
        Column name(s) of text data that you want to stem

    regexp : str, optional
        Regex pattern of punctuation to be removed.

    exceptions : list, optional
        List of punctuation to remove.

    new_col_name : str, optional
        New column name to be created when applying this technique, by default `_rem_punct`

    Returns
    -------
    Dataframe, *Dataframe
        Transformed dataframe with the new column.

    Returns 2 Dataframes if x_test is provided. 
    """

    delete_punct = set(string.punctuation) - set(exceptions)
    tokenizer = RegexpTokenizer(regexp)

    for col in list_of_cols:
        if new_col_name.startswith("_"):
            new_col_name = col + new_col_name

        if not regexp:
            x_train[new_col_name] = list(
                map(
                    lambda x: "".join(
                        [letter for letter in x if letter not in delete_punct]
                    ),
                    x_train[col],
                )
            )

            if x_test is not None:
                x_test[new_col_name] = list(
                    map(
                        lambda x: "".join(
                            [letter for letter in x if letter not in delete_punct]
                        ),
                        x_test[col],
                    )
                )
        else:
            x_train[new_col_name] = list(
                map(lambda x: " ".join(tokenizer.tokenize(x)), x_train[col])
            )

            if x_test is not None:
                x_test[new_col_name] = list(
                    map(lambda x: " ".join(tokenizer.tokenize(x)), x_test[col])
                )

    return x_train, x_test


def remove_numbers(
    x_train, x_test=None, list_of_cols=[], new_col_name="_rem_num",
):
    """
    Removes numbers from text.

    Parameters
    ----------
    x_train : DataFrame
        Dataset

    x_test : DataFrame
        Testing dataset, by default None

    list_of_cols : list, optional
        Column name(s) of text data


    new_col_name : str, optional
        New column name to be created when applying this technique, by default `_rem_num`

    Returns
    -------
    Dataframe, *Dataframe
        Transformed dataframe with the new column.

    Returns 2 Dataframes if x_test is provided. 
    """

    for col in list_of_cols:
        if new_col_name.startswith("_"):
            new_col_name = col + new_col_name

        x_train[new_col_name] = pd.Series(
            map(
                lambda x: str.translate(x, str.maketrans("", "", "0123456789")),
                x_train[col],
            )
        )

        if x_test is not None:
            x_test[new_col_name] = pd.Series(
                map(
                    lambda x: str.translate(x, str.maketrans("", "", "0123456789")),
                    x_test[col],
                )
            )

        return x_train, x_test


def process_text(
    corpus, lower=True, punctuation=True, stopwords=True, stemmer=True, numbers=True
):
    """
    Function that takes text and does the following:

      - Casts it to lowercase
      - Removes punctuation
      - Removes stopwords
      - Stems the text
      - Removes any numerical values
    
    Parameters
    ----------
    corpus : str
        Text
    
    lower : bool, optional
        True to cast all text to lowercase, by default True
    
    punctuation : bool, optional
        True to remove punctuation, by default True
    
    stopwords : bool, optional
        True to remove stop words, by default True
    
    stemmer : bool, optional
        True to stem the data, by default True

    numbers : bool, optional
        True to remove any numbers, by default True
    
    Returns
    -------
    str
        Normalized text
    """

    import nltk

    transformed_corpus = ""

    if lower:
        corpus = corpus.lower()

    for token in word_tokenize(corpus):

        if punctuation:
            if token in string.punctuation:
                continue

            token = token.translate(str.maketrans("", "", string.punctuation))

        if numbers:
            token = token.translate(str.maketrans("", "", "0123456789"))

        if stopwords:
            stop_words = nltk.corpus.stopwords.words("english")
            if token in stop_words:
                continue

        if stemmer:
            stem = SnowballStemmer("english")
            token = stem.stem(token)

        transformed_corpus += token + " "

    return transformed_corpus.strip()


def _apply_text_method(text_data, transformer=None):
    """
    Applies a text based method to the given data, for example
    a Lemmatizer, Stemmer, etc.

    Parameters
    ----------
    text_data : str
        Text data to transform
        
    transformer : Transformation object, optional
        trasnformer applied on the data, for example
        lemmatizer, stemmer, etc. , by default None

    Returns
    -------
    str
        Transformed data
    """

    transformed_text_data = ""

    for word in text_data.split():
        transformed_text_data += f"{transformer(word)} "

    return transformed_text_data.strip()



import itertools
import math
import os
import warnings
from collections import OrderedDict
from itertools import compress

import xgboost as xgb
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import sklearn
from IPython.display import HTML, SVG, display

from aethos.config import IMAGE_DIR
from aethos.config.config import _global_config
from aethos.feature_engineering.util import sklearn_dim_reduction
from aethos.modelling.model_explanation import MSFTInterpret, Shap
from aethos.modelling.util import _make_img_project_dir, to_pickle, track_artifacts
from aethos.templates.template_generator import TemplateGenerator as tg
from aethos.visualizations.visualizations import Visualizations
from aethos.stats.stats import Stats


class ModelBase(Visualizations, Stats):

    # TODO: Add more SHAP use cases

    def __init__(self, model_object, model, model_name, **kwargs):

        from aethos.modelling.constants import (
            PROBLEM_TYPE,
            SHAP_LEARNERS,
        )

        self.model = model
        self.model_name = model_name
        self.x_train = model_object.x_train
        self.x_test = model_object.x_test
        self.report = model_object.report

        shap_values = kwargs.pop("shap_values", None)

        if isinstance(self, ClassificationModel) or isinstance(self, RegressionModel):
            self.shap = Shap(
                self.model,
                self.model_name,
                self.x_train,
                self.x_test,
                self.y_test,
                SHAP_LEARNERS[type(self.model)],
                shap_values,
            )
            self.interpret = MSFTInterpret(
                self.model,
                self.x_train,
                self.x_test,
                self.y_test,
                PROBLEM_TYPE[type(self.model)],
            )
        else:
            self.shap = None
            self.interpret = None

        for method in dir(self.model):
            try:
                if not method.startswith("_") and not method.startswith("predict"):
                    self.__setattr__(method, getattr(self.model, method))

            except AttributeError as e:
                continue

    def model_weights(self):
        """
        Prints and logs all the features ranked by importance from most to least important.
        
        Returns
        -------
        dict
            Dictionary of features and their corresponding weights
        
        Raises
        ------
        AttributeError
            If model does not have coefficients to display

        Examples
        --------
        >>> m = model.LogisticRegression()
        >>> m.model_weights()
        """

        report_strings = []

        try:
            model_dict = dict(zip(self.features, self.model.coef_.flatten()))
        except Exception as e:
            raise AttributeError("Model does not have coefficients to view.")

        sorted_features = OrderedDict(
            sorted(model_dict.items(), key=lambda kv: abs(kv[1]), reverse=True)
        )

        for feature, weight in sorted_features.items():
            report_string = "\t{} : {:.2f}".format(feature, weight)
            report_strings.append(report_string)

            print(report_string.strip())

        if self.report:
            self.report.log("Features ranked from most to least important:\n")
            self.report.write_contents("\n".join(report_strings))

        return sorted_features

    def summary_plot(self, output_file="", **summaryplot_kwargs):
        """
        Create a SHAP summary plot, colored by feature values when they are provided.

        For a list of all kwargs please see the Shap documentation : https://shap.readthedocs.io/en/latest/#plots

        Parameters
        ----------
        output_file: str
            Output file name including extension (.png, .jpg, etc.) to save image as.

        max_display : int
            How many top features to include in the plot (default is 20, or 7 for interaction plots), by default None
            
        plot_type : "dot" (default for single output), "bar" (default for multi-output), "violin", or "compact_dot"
            What type of summary plot to produce. Note that "compact_dot" is only used for SHAP interaction values.

        color : str or matplotlib.colors.ColorMap 
            Color spectrum used to draw the plot lines. If str, a registered matplotlib color name is assumed.

        axis_color : str or int 
            Color used to draw plot axes.

        title : str 
            Title of the plot.

        alpha : float 
            Alpha blending value in [0, 1] used to draw plot lines.

        show : bool 
            Whether to automatically display the plot.

        sort : bool
            Whether to sort features by importance, by default True

        color_bar : bool 
            Whether to draw the color bar.

        auto_size_plot : bool 
            Whether to automatically size the matplotlib plot to fit the number of features displayed. If False, specify the plot size using matplotlib before calling this function.

        layered_violin_max_num_bins : int
            Max number of bins, by default 20

        **summaryplot_kwargs
            For more info see https://shap.readthedocs.io/en/latest/#plots

        Examples
        --------
        >>> m = model.LogisticRegression()
        >>> m.summary_plot()
        """

        if self.shap is None:
            raise NotImplementedError(
                f"SHAP is not implemented yet for {str(type(self))}"
            )

        self.shap.summary_plot(output_file=output_file, **summaryplot_kwargs)

        if output_file and self.report:
            self.report.write_image(output_file)

        if _global_config["track_experiments"]:  # pragma: no cover
            track_artifacts(self.run_id, self.model_name)

    def decision_plot(
        self,
        num_samples=0.6,
        sample_no=None,
        highlight_misclassified=False,
        output_file="",
        **decisionplot_kwargs,
    ):
        """
        Visualize model decisions using cumulative SHAP values.
        
        Each colored line in the plot represents the model prediction for a single observation. 
        
        Note that plotting too many samples at once can make the plot unintelligible.

        When is a decision plot useful:        
            - Show a large number of feature effects clearly.
            
            - Visualize multioutput predictions.
            
            - Display the cumulative effect of interactions.
            
            - Explore feature effects for a range of feature values.
            
            - Identify outliers.
            
            - Identify typical prediction paths.
            
            - Compare and contrast predictions for several models.

        Explanation:
            - The plot is centered on the x-axis at the models expected value.

            - All SHAP values are relative to the model's expected value like a linear model's effects are relative to the intercept.
            
            - The y-axis lists the model's features. By default, the features are ordered by descending importance. 
            
            - The importance is calculated over the observations plotted. This is usually different than the importance ordering for the entire dataset. In addition to feature importance ordering, the decision plot also supports hierarchical cluster feature ordering and user-defined feature ordering.
            
            - Each observation's prediction is represented by a colored line. 
            
            - At the top of the plot, each line strikes the x-axis at its corresponding observation's predicted value. This value determines the color of the line on a spectrum.
            
            - Moving from the bottom of the plot to the top, SHAP values for each feature are added to the model's base value. This shows how each feature contributes to the overall prediction.
            
            - At the bottom of the plot, the observations converge at the models expected value.

        Parameters
        ----------
        output_file: str
            Output file name including extension (.png, .jpg, etc.) to save image as.

        num_samples : int, float, or 'all', optional
            Number of samples to display, if less than 1 it will treat it as a percentage, 'all' will include all samples
            , by default 0.6

        sample_no : int, optional
            Sample number to isolate and analyze, if provided it overrides num_samples, by default None

        highlight_misclassified : bool, optional
            True to highlight the misclassified results, by default False

        feature_order : str or None or list or numpy.ndarray
            Any of "importance" (the default), "hclust" (hierarchical clustering), "none", or a list/array of indices.
            hclust is useful for finding outliers.

        feature_display_range: slice or range
            The slice or range of features to plot after ordering features by feature_order. A step of 1 or None will display the features in ascending order. A step of -1 will display the features in descending order. If feature_display_range=None, slice(-1, -21, -1) is used (i.e. show the last 20 features in descending order). If shap_values contains interaction values, the number of features is automatically expanded to include all possible interactions: N(N + 1)/2 where N = shap_values.shape[1].

        highlight : Any 
            Specify which observations to draw in a different line style. All numpy indexing methods are supported. For example, list of integer indices, or a bool array.

        link : str 
            Use "identity" or "logit" to specify the transformation used for the x-axis. The "logit" link transforms log-odds into probabilities.

        plot_color : str or matplotlib.colors.ColorMap 
            Color spectrum used to draw the plot lines. If str, a registered matplotlib color name is assumed.

        axis_color : str or int 
            Color used to draw plot axes.

        y_demarc_color : str or int 
            Color used to draw feature demarcation lines on the y-axis.

        alpha : float 
            Alpha blending value in [0, 1] used to draw plot lines.

        color_bar : bool 
            Whether to draw the color bar.

        auto_size_plot : bool 
            Whether to automatically size the matplotlib plot to fit the number of features displayed. If False, specify the plot size using matplotlib before calling this function.

        title : str 
            Title of the plot.

        xlim: tuple[float, float] 
            The extents of the x-axis (e.g. (-1.0, 1.0)). If not specified, the limits are determined by the maximum/minimum predictions centered around base_value when link='identity'. When link='logit', the x-axis extents are (0, 1) centered at 0.5. x_lim values are not transformed by the link function. This argument is provided to simplify producing multiple plots on the same scale for comparison.

        show : bool 
            Whether to automatically display the plot.

        return_objects : bool 
            Whether to return a DecisionPlotResult object containing various plotting features. This can be used to generate multiple decision plots using the same feature ordering and scale, by default True.

        ignore_warnings : bool 
            Plotting many data points or too many features at a time may be slow, or may create very large plots. Set this argument to True to override hard-coded limits that prevent plotting large amounts of data.

        new_base_value : float 
            SHAP values are relative to a base value; by default, the expected value of the model's raw predictions. Use new_base_value to shift the base value to an arbitrary value (e.g. the cutoff point for a binary classification task).

        legend_labels : list of str 
            List of legend labels. If None, legend will not be shown.

        legend_location : str 
            Legend location. Any of "best", "upper right", "upper left", "lower left", "lower right", "right", "center left", "center right", "lower center", "upper center", "center".
    
        Returns
        -------
        DecisionPlotResult 
            If return_objects=True (the default). Returns None otherwise.

        Examples
        --------
        >>> # Plot two decision plots using the same feature order and x-axis.
        >>> m = model.LogisticRegression()
        >>> r = m.decision_plot()
        >>> m.decision_plot(no_sample=42, feature_order=r.feature_idx, xlim=r.xlim)
        """

        if self.shap is None:
            raise NotImplementedError(
                f"SHAP is not implemented yet for {str(type(self))}"
            )

        if highlight_misclassified:
            if not any(self.shap.misclassified_values):
                raise AttributeError("There are no misclassified values!")

            decisionplot_kwargs["highlight"] = self.shap.misclassified_values

        dp = self.shap.decision_plot(
            num_samples, sample_no, output_file=output_file, **decisionplot_kwargs
        )

        if output_file and self.report:
            self.report.write_image(output_file)

        if _global_config["track_experiments"]:  # pragma: no cover
            track_artifacts(self.run_id, self.model_name)

        return dp

    def force_plot(
        self, sample_no=None, misclassified=False, output_file="", **forceplot_kwargs
    ):
        """
        Visualize the given SHAP values with an additive force layout
        
        Parameters
        ----------
        sample_no : int, optional
            Sample number to isolate and analyze, by default None

        misclassified : bool, optional
            True to only show the misclassified results, by default False

        output_file: str
            Output file name including extension (.png, .jpg, etc.) to save image as.

        link : "identity" or "logit"
            The transformation used when drawing the tick mark labels. Using logit will change log-odds numbers
            into probabilities. 

        matplotlib : bool
            Whether to use the default Javascript output, or the (less developed) matplotlib output. Using matplotlib
            can be helpful in scenarios where rendering Javascript/HTML is inconvenient. 
        
        Examples
        --------
        >>> m = model.LogisticRegression()
        >>> m.force_plot() # The entire test dataset
        >>> m.forceplot(no_sample=1, misclassified=True) # Analyze the first misclassified result
        """

        if self.shap is None:
            raise NotImplementedError(
                f"SHAP is not implemented yet for {str(type(self))}"
            )

        if misclassified:
            if not any(self.shap.misclassified_values):
                raise AttributeError("There are no misclassified values!")

            forceplot_kwargs["shap_values"] = self.shap.shap_values[
                self.shap.misclassified_values
            ]

        fp = self.shap.force_plot(
            sample_no, output_file=output_file, **forceplot_kwargs
        )

        if output_file and self.report:
            self.report.write_image(output_file)

        if _global_config["track_experiments"]:  # pragma: no cover
            track_artifacts(self.run_id, self.model_name)

        return fp

    def dependence_plot(
        self, feature: str, interaction="auto", output_file="", **dependenceplot_kwargs
    ):
        """
        A dependence plot is a scatter plot that shows the effect a single feature has on the predictions made by the mode.

        Explanation:
            - Each dot is a single prediction (row) from the dataset.
        
            - The x-axis is the value of the feature (from the X matrix).

            - The y-axis is the SHAP value for that feature, which represents how much knowing that feature's value changes the output of the model for that sample's prediction.

            - The color corresponds to a second feature that may have an interaction effect with the feature we are plotting (by default this second feature is chosen automatically). 
            
            - If an interaction effect is present between this other feature and the feature we are plotting it will show up as a distinct vertical pattern of coloring. 
        
        Parameters
        ----------
        feature : str
            Feature who's impact on the model you want to analyze

        interaction : "auto", None, int, or string
            The index of the feature used to color the plot. The name of a feature can also be passed as a string. If "auto" then shap.common.approximate_interactions is used to pick what seems to be the strongest interaction (note that to find to true stongest interaction you need to compute the SHAP interaction values).
        
        output_file: str
            Output file name including extension (.png, .jpg, etc.) to save image as.

        x_jitter : float (0 - 1)
            Adds random jitter to feature values. May increase plot readability when feature is discrete.

        alpha : float
            The transparency of the data points (between 0 and 1). This can be useful to the show density of the data points when using a large dataset.

        xmin : float or string
            Represents the lower bound of the plot's x-axis. It can be a string of the format "percentile(float)" to denote that percentile of the feature's value used on the x-axis.

        xmax : float or string
            Represents the upper bound of the plot's x-axis. It can be a string of the format "percentile(float)" to denote that percentile of the feature's value used on the x-axis.

        ax : matplotlib Axes object
            Optionally specify an existing matplotlib Axes object, into which the plot will be placed. In this case we do not create a Figure, otherwise we do.

        cmap : str or matplotlib.colors.ColorMap 
            Color spectrum used to draw the plot lines. If str, a registered matplotlib color name is assumed.

        Examples
        --------
        >>> m = model.LogisticRegression()
        >>> m.dependence_plot()
        """

        if self.shap is None:
            raise NotImplementedError(
                f"SHAP is not implemented yet for {str(type(self))}"
            )

        dp = self.shap.dependence_plot(
            feature, interaction, output_file=output_file, **dependenceplot_kwargs
        )

        if output_file and self.report:
            self.report.write_image(output_file)

        if _global_config["track_experiments"]:  # pragma: no cover
            track_artifacts(self.run_id, self.model_name)

        return dp

    def shap_get_misclassified_index(self):
        """
        Prints the sample numbers of misclassified samples.

        Examples
        --------
        >>> m = model.LogisticRegression()
        >>> m.shap_get_misclassified_index()
        """

        sample_list = list(
            compress(
                range(len(self.shap.misclassified_values)),
                self.shap.misclassified_values,
            )
        )

        return sample_list

    def interpret_model(self, show=True):  # pragma: no cover
        """
        Displays a dashboard interpreting your model's performance, behaviour and individual predictions.

        If you have run any other `interpret` functions, they will be included in the dashboard, otherwise all the other intrepretable methods will be included in the dashboard.

        Examples
        --------
        >>> m = model.LogisticRegression()
        >>> m.interpret_model()
        """

        warnings.simplefilter("ignore")

        if isinstance(self.model, xgb.XGBModel):
            return "Using MSFT interpret is currently unsupported with XGBoost."

        if show:
            self.interpret.create_dashboard()

    def interpret_model_performance(
        self, method="all", predictions="default", show=True, **interpret_kwargs
    ):
        """
        Plots an interpretable display of your model based off a performance metric.

        Can either be 'ROC' or 'PR' for precision, recall for classification problems.

        Can be 'regperf' for regression problems.

        If 'all' a dashboard is displayed with the corresponding explainers for the problem type.

        ROC: Receiver Operator Characteristic
        PR: Precision Recall
        regperf: RegeressionPerf
        
        Parameters
        ----------
        method : str
            Performance metric, either 'all', 'roc' or 'PR', by default 'all'

        predictions : str, optional
            Prediction type, can either be 'default' (.predict) or 'probability' if the model can predict probabilities, by default 'default'

        show : bool, optional 
            False to not display the plot, by default True

        Examples
        --------
        >>> m = model.LogisticRegression()
        >>> m.interpret_model_performance()
        """

        import interpret
        from aethos.modelling.constants import INTERPRET_EXPLAINERS

        warnings.simplefilter("ignore")

        if isinstance(self.model, xgb.XGBModel):  # pragma: no cover
            return "Using MSFT interpret is currently unsupported with XGBoost."

        dashboard = []

        if method == "all":
            for explainer in INTERPRET_EXPLAINERS["problem"][self.interpret.problem]:
                dashboard.append(
                    self.interpret.blackbox_show_performance(
                        method=explainer,
                        predictions=predictions,
                        show=False,
                        **interpret_kwargs,
                    )
                )

            if show:
                interpret.show(dashboard)
        else:
            self.interpret.blackbox_show_performance(
                method=method, predictions=predictions, show=show, **interpret_kwargs
            )

    def interpret_model_predictions(
        self,
        num_samples=0.25,
        sample_no=None,
        method="all",
        predictions="default",
        show=True,
        **interpret_kwargs,
    ):
        """
        Plots an interpretable display that explains individual predictions of your model.

        Supported explainers are either 'lime' or 'shap'.

        If 'all' a dashboard is displayed with morris and dependence analysis displayed.
        
        Parameters
        ----------
        num_samples : int, float, or 'all', optional
            Number of samples to display, if less than 1 it will treat it as a percentage, 'all' will include all samples
            , by default 0.25

        sample_no : int, optional
            Sample number to isolate and analyze, if provided it overrides num_samples, by default None

        method : str, optional
            Explainer type, can either be 'all', 'lime', or 'shap', by default 'all'

        predictions : str, optional
            Prediction type, can either be 'default' (.predict) or 'probability' if the model can predict probabilities, by default 'default'

        show : bool, optional 
            False to not display the plot, by default True

        Examples
        --------
        >>> m = model.LogisticRegression()
        >>> m.interpret_model_predictions()
        """

        import interpret
        from aethos.modelling.constants import INTERPRET_EXPLAINERS

        warnings.simplefilter("ignore")

        if isinstance(self.model, xgb.XGBModel):  # pragma: no cover
            return "Using MSFT interpret is currently unsupported with XGBoost."

        dashboard = []

        if method == "all":
            for explainer in INTERPRET_EXPLAINERS["local"]:
                dashboard.append(
                    self.interpret.blackbox_local_explanation(
                        num_samples=num_samples,
                        sample_no=sample_no,
                        method=explainer,
                        predictions=predictions,
                        show=False,
                        **interpret_kwargs,
                    )
                )

            if show:
                interpret.show(dashboard)
        else:
            self.interpret.blackbox_local_explanation(
                num_samples=num_samples,
                sample_no=sample_no,
                method=method,
                predictions=predictions,
                show=show,
                **interpret_kwargs,
            )

    def interpret_model_behavior(
        self, method="all", predictions="default", show=True, **interpret_kwargs
    ):
        """
        Provides an interpretable summary of your models behaviour based off an explainer.

        Can either be 'morris' or 'dependence' for Partial Dependence.
        
        If 'all' a dashboard is displayed with morris and dependence analysis displayed.
        
        Parameters
        ----------
        method : str, optional
            Explainer type, can either be 'all', 'morris' or 'dependence', by default 'all'

        predictions : str, optional
            Prediction type, can either be 'default' (.predict) or 'probability' if the model can predict probabilities, by default 'default'

        show : bool, optional 
            False to not display the plot, by default True

        Examples
        --------
        >>> m = model.LogisticRegression()
        >>> m.interpret_model_behavior()
        """

        import interpret
        from aethos.modelling.constants import INTERPRET_EXPLAINERS

        warnings.simplefilter("ignore")

        if isinstance(self.model, xgb.XGBModel):  # pragma: no cover
            return "Using MSFT interpret is currently unsupported with XGBoost."

        dashboard = []

        if method == "all":
            for explainer in INTERPRET_EXPLAINERS["global"]:
                dashboard.append(
                    self.interpret.blackbox_global_explanation(
                        method=explainer,
                        predictions=predictions,
                        show=False,
                        **interpret_kwargs,
                    )
                )

            if show:
                interpret.show(dashboard)
        else:
            self.interpret.blackbox_global_explanation(
                method=method, predictions=predictions, show=show, **interpret_kwargs
            )

    def view_tree(self, tree_num=0, output_file=None, **kwargs):
        """
        Plot decision trees.
        
        Parameters
        ----------
        tree_num: int, optional
            For ensemble, boosting, and stacking methods - the tree number to plot, by default 0

        output_file : str, optional
            Name of the file including extension, by default None

        Examples
        --------
        >>> m = model.DecisionTreeClassifier()
        >>> m.view_tree()
        >>> m = model.XGBoostClassifier()
        >>> m.view_tree(2)
        """

        import catboost as cb
        import lightgbm as lgb
        from graphviz import Source

        if hasattr(self, "classes"):
            classes = self.classes
        else:
            classes = None

        if isinstance(self.model, sklearn.tree.BaseDecisionTree):
            graph = Source(
                sklearn.tree.export_graphviz(
                    self.model,
                    out_file=None,
                    feature_names=self.features,
                    class_names=classes,
                    rounded=True,
                    precision=True,
                    filled=True,
                )
            )

            display(SVG(graph.pipe(format="svg")))

        elif isinstance(self.model, xgb.XGBModel):
            return xgb.plot_tree(self.model)

        elif isinstance(self.model, lgb.sklearn.LGBMModel):
            return lgb.plot_tree(self.model)

        elif isinstance(self.model, cb.CatBoost):
            return self.model.plot_tree(tree_idx=tree_num, pool=self.pool)

        elif isinstance(self.model, sklearn.ensemble.BaseEnsemble):
            estimator = self.model.estimators_[tree_num]

            graph = Source(
                sklearn.tree.export_graphviz(
                    estimator,
                    out_file=None,
                    feature_names=self.features,
                    class_names=classes,
                    rounded=True,
                    precision=True,
                    filled=True,
                )
            )

            display(SVG(graph.pipe(format="svg")))
        else:
            raise NotImplementedError(
                f"Model {str(self.model)} cannot be viewed as a tree"
            )

    def to_pickle(self):
        """
        Writes model to a pickle file.

        Examples
        --------
        >>> m = Model(df)
        >>> m_results = m.LogisticRegression()
        >>> m_results.to_pickle()
        """

        to_pickle(self.model, self.model_name)

    def to_service(self, project_name: str):
        """
        Creates an app.py, requirements.txt and Dockerfile in `~/.aethos/projects` and the necessary folder structure
        to run the model as a microservice.
        
        Parameters
        ----------
        project_name : str
            Name of the project that you want to create.

        Examples
        --------
        >>> m = Model(df)
        >>> m_results = m.LogisticRegression()
        >>> m_results.to_service('your_proj_name')
        """

        to_pickle(self.model, self.model_name, project=True, project_name=project_name)
        tg.generate_service(project_name, f"{self.model_name}.pkl", self.model)

        print("To run:")
        print("\tdocker build -t `image_name` ./")
        print("\tdocker run -d --name `container_name` -p `port_num`:80 `image_name`")


class TextModel(ModelBase):
    def __init__(self, model_object, model, model_name, **kwargs):

        super().__init__(model_object, model, model_name)

        self.result_data = model_object.x_train_results

        # LDA dependant variables
        self.corpus = kwargs.pop("corpus", None)
        self.id2word = kwargs.pop("id2word", None)

    def view(self, original_text, model_output):
        """
        View the original text and the model output in a more user friendly format
        
        Parameters
        ----------
        original_text : str
            Column name of the original text

        model_output : str
            Column name of the model text

        Examples
        --------
        >>> m = model.LDA()
        >>> m.view('original_text_col_name', 'model_output_col_name')
        """

        results = self.result_data[[original_text, model_output]]

        display(HTML(results.to_html()))

    def view_topics(self, num_topics=10, **kwargs):
        """
        View topics from topic modelling model.
        
        Parameters
        ----------
        num_topics : int, optional
            Number of topics to view, by default 10

        Returns
        --------
        str
            String representation of topics and probabilities

        Examples
        --------
        >>> m = model.LDA()
        >>> m.view_topics()
        """

        return self.model.show_topics(num_topics=num_topics, **kwargs)

    def view_topic(self, topic_num: int, **kwargs):
        """
        View a specific topic from topic modelling model.
        
        Parameters
        ----------
        topic_num : int

        Returns
        --------
        str
            String representation of topic and probabilities

        Examples
        --------
        >>> m = model.LDA()
        >>> m.view_topic(1)
        """

        return self.model.show_topic(topicid=topic_num, **kwargs)

    def visualize_topics(self, **kwargs):  # pragma: no cover
        """
        Visualize topics using pyLDAvis.

        Parameters
        ----------
        R : int
            The number of terms to display in the barcharts of the visualization. Default is 30. Recommended to be roughly between 10 and 50.

        lambda_step : float, between 0 and 1
            Determines the interstep distance in the grid of lambda values over which to iterate when computing relevance. Default is 0.01. Recommended to be between 0.01 and 0.1.

        mds : function or {'tsne', 'mmds}
            A function that takes topic_term_dists as an input and outputs a n_topics by 2 distance matrix.
            The output approximates the distance between topics. See js_PCoA() for details on the default function.
            A string representation currently accepts pcoa (or upper case variant), mmds (or upper case variant) and tsne (or upper case variant), if sklearn package is installed for the latter two.

        n_jobs : int
            The number of cores to be used to do the computations. The regular joblib conventions are followed so -1, which is the default, will use all cores.

        plot_opts : dict, with keys ‘xlab’ and ylab
            Dictionary of plotting options, right now only used for the axis labels.

        sort_topics : bool
            Sort topics by topic proportion (percentage of tokens covered).
            Set to false to keep original topic order.

        Examples
        --------
        >>> m = model.LDA()
        >>> m.visualize_topics()
        """

        import pyLDAvis
        import pyLDAvis.gensim

        pyLDAvis.enable_notebook()

        return pyLDAvis.gensim.prepare(self.model, self.corpus, self.id2word, **kwargs)

    def coherence_score(self, col_name):
        """
        Displays the coherence score of the topic model.

        For more info on topic coherence: https://rare-technologies.com/what-is-topic-coherence/ 
        
        Parameters
        ----------
        col_name : str
            Column name that was used as input for the LDA model

        Examples
        --------
        >>> m = model.LDA()
        >>> m.coherence_score()
        """

        import gensim
        import plotly.graph_objects as go

        texts = self.result_data[col_name].tolist()

        coherence_model_lda = gensim.models.CoherenceModel(
            model=self.model, texts=texts, dictionary=self.id2word, coherence="c_v"
        )
        coherence_lda = coherence_model_lda.get_coherence()

        fig = go.Figure(
            go.Indicator(
                domain={"x": [0, 1], "y": [0, 1]},
                value=coherence_lda,
                mode="number",
                title={"text": "Coherence Score"},
            )
        )

        fig.show()

    def model_perplexity(self):
        """
        Displays the model perplexity of the topic model.

        Perplexity is a measurement of how well a probability distribution or probability model predicts a sample. It may be used to compare probability models.
        
        A low perplexity indicates the probability distribution is good at predicting the sample.

        Examples
        --------
        >>> m = model.LDA()
        >>> m.model_perplexity()
        """

        import plotly.graph_objects as go

        fig = go.Figure(
            go.Indicator(
                domain={"x": [0, 1], "y": [0, 1]},
                value=self.model.log_perplexity(self.corpus),
                mode="number",
                title={"text": "Model Perplexity"},
            )
        )

        fig.show()

        print("Note: The lower the better.")


class UnsupervisedModel(ModelBase):
    def __init__(self, model_object, model_name, model, cluster_col):

        super().__init__(model_object, model, model_name)

        self.cluster_col = cluster_col

        self.x_train[self.cluster_col] = model_object.x_train_results[self.cluster_col]

        if self.x_test is not None:
            self.x_test[self.cluster_col] = model_object.x_test_results[
                self.cluster_col
            ]

    def filter_cluster(self, cluster_no: int):
        """
        Filters data by a cluster number for analysis.
        
        Parameters
        ----------
        cluster_no : int
            Cluster number to filter by
        
        Returns
        -------
        Dataframe
            Filtered data or test dataframe

        Examples
        --------
        >>> m = model.KMeans()
        >>> m.filter_cluster(1)
        """

        if self.x_test is None:
            return self.x_train[self.x_train[self.cluster_col] == cluster_no]
        else:
            return self.x_test[self.x_test[self.cluster_col] == cluster_no]

    def plot_clusters(self, dim=2, reduce="pca", output_file="", **kwargs):
        """
        Plots the clusters in either 2d or 3d space with each cluster point highlighted
        as a different colour.

        For 2d plotting options, see:
        
            https://bokeh.pydata.org/en/latest/docs/reference/plotting.html#bokeh.plotting.figure.Figure.scatter

            https://bokeh.pydata.org/en/latest/docs/user_guide/styling.html#userguide-styling-line-properties 

        For 3d plotting options, see:

            https://www.plotly.express/plotly_express/#plotly_express.scatter_3d
            
        Parameters
        ----------
        dim : 2 or 3, optional
            Dimension of the plot, either 2 for 2d, 3 for 3d, by default 2

        reduce : str {'pca', 'tvsd', 'lle', 'tsne'}, optional
            Dimension reduction strategy i.e. pca, by default "pca"

        output_file: str
            Output file name including extension (.png, .jpg, etc.) to save image as.

        Examples
        --------
        >>> m = model.KMeans()
        >>> m.plot_clusters()
        >>> m.plot_clusters(dim=3)
        """

        dataset = self.x_test if self.x_test is not None else self.x_train
        output_file_path = (
            os.path.join(self.model_name, output_file) if output_file else output_file
        )

        self.plot_dim_reduction(
            category=self.cluster_col,
            dim=dim,
            algo=reduce,
            output_file=output_file_path,
            **kwargs,
        )

        if _global_config["track_experiments"]:  # pragma: no cover
            track_artifacts(self.run_id, self.model_name)


class ClassificationModel(ModelBase):
    def __init__(
        self,
        model_object,
        model_name,
        model,
        predictions_col,
        shap_values=None,
        pool=None,
        run_id=None,
    ):

        self.y_train = model_object.y_train
        self.y_test = (
            model_object.y_test
            if model_object.x_test is not None
            else model_object.y_train
        )

        super().__init__(model_object, model, model_name, shap_values=shap_values)

        self.probabilities = None
        self.target_mapping = model_object.target_mapping
        self.multiclass = len(np.unique(list(self.y_train) + list(self.y_test))) > 2

        self.y_pred = (
            model_object.x_train_results[predictions_col]
            if self.x_test is None
            else model_object.x_test_results[predictions_col]
        )

        if self.report:
            self.report.write_header(f"Analyzing Model {self.model_name.upper()}: ")

        if self.target_mapping is None:
            self.classes = [
                str(item) for item in np.unique(list(self.y_train) + list(self.y_test))
            ]
        else:
            self.classes = [str(item) for item in self.target_mapping.values()]

        if hasattr(model, "predict_proba"):
            self.probabilities = model.predict_proba(model_object.x_test)

        self.features = self.x_test.columns
        self.pool = pool
        self.run_id = run_id

    def accuracy(self, **kwargs):
        """
        It measures how many observations, both positive and negative, were correctly classified.
        
        Returns
        -------
        float
            Accuracy

        Examples
        --------
        >>> m = model.LogisticRegression()
        >>> m.accuracy()
        """

        return sklearn.metrics.accuracy_score(self.y_test, self.y_pred, **kwargs)

    def balanced_accuracy(self, **kwargs):
        """
        The balanced accuracy in binary and multiclass classification problems to deal with imbalanced datasets. It is defined as the average of recall obtained on each class.

        The best value is 1 and the worst value is 0 when adjusted=False.
        
        Returns
        -------
        float
            Balanced accuracy

        Examples
        --------
        >>> m = model.LogisticRegression()
        >>> m.balanced_accuracy()
        """

        return sklearn.metrics.balanced_accuracy_score(
            self.y_test, self.y_pred, **kwargs
        )

    def average_precision(self, **kwargs):
        """
        AP summarizes a precision-recall curve as the weighted mean of precisions achieved at each threshold,
        with the increase in recall from the previous threshold used as the weight
        
        Returns
        -------
        float
            Average Precision Score

        Examples
        --------
        >>> m = model.LogisticRegression()
        >>> m.average_precision()
        """

        if hasattr(self.model, "decision_function"):
            return sklearn.metrics.average_precision_score(
                self.y_test, self.model.decision_function(self.x_test), **kwargs
            )
        else:
            return np.nan

    def roc_auc(self, **kwargs):
        """
        This metric tells us that this metric shows how good at ranking predictions your model is.
        It tells you what is the probability that a randomly chosen positive instance is ranked higher than a randomly chosen negative instance.
        
        Returns
        -------
        float
            ROC AUC Score

        Examples
        --------
        >>> m = model.LogisticRegression()
        >>> m.roc_auc()
        """

        multi_class = kwargs.pop("multi_class", "ovr")

        if self.multiclass:
            roc_auc = sklearn.metrics.roc_auc_score(
                self.y_test, self.probabilities, multi_class=multi_class, **kwargs
            )
        else:
            roc_auc = sklearn.metrics.roc_auc_score(self.y_test, self.y_pred, **kwargs)

        return roc_auc

    def zero_one_loss(self, **kwargs):
        """
        Return the fraction of misclassifications (float), else it returns the number of misclassifications (int).
        
        The best performance is 0.
        
        Returns
        -------
        float
            Zero one loss

        Examples
        --------
        >>> m = model.LogisticRegression()
        >>> m.zero_one_loss()
        """

        return sklearn.metrics.zero_one_loss(self.y_test, self.y_pred, **kwargs)

    def recall(self, **kwargs):
        """
        The recall is the ratio tp / (tp + fn) where tp is the number of true positives and fn the number of false negatives. 
        
        The recall is intuitively the ability of the classifier to find all the positive samples.

        The best value is 1 and the worst value is 0.
        
        Returns
        -------
        float
            Recall

        Examples
        --------
        >>> m = model.LogisticRegression()
        >>> m.recall()
        """

        avg = kwargs.pop("average", "macro")

        if self.multiclass:
            return sklearn.metrics.recall_score(
                self.y_test, self.y_pred, average=avg, **kwargs
            )
        else:
            return sklearn.metrics.recall_score(self.y_test, self.y_pred, **kwargs)

    def precision(self, **kwargs):
        """
        The precision is the ratio tp / (tp + fp) where tp is the number of true positives and fp the number of false positives.
        
        The precision is intuitively the ability of the classifier not to label as positive a sample that is negative.

        The best value is 1 and the worst value is 0.
        
        Returns
        -------
        float
            Precision

        Examples
        --------
        >>> m = model.LogisticRegression()
        >>> m.precision()
        """

        avg = kwargs.pop("average", "macro")

        if self.multiclass:
            return sklearn.metrics.precision_score(
                self.y_test, self.y_pred, average=avg, **kwargs
            )
        else:
            return sklearn.metrics.precision_score(self.y_test, self.y_pred, **kwargs)

    def matthews_corr_coef(self, **kwargs):
        """
        The Matthews correlation coefficient is used in machine learning as a measure of the quality of binary and multiclass classifications.
        It takes into account true and false positives and negatives and is generally regarded as a balanced measure which can be used even if the classes are of very different sizes.
        The MCC is in essence a correlation coefficient value between -1 and +1. 
        A coefficient of +1 represents a perfect prediction, 0 an average random prediction and -1 an inverse prediction.
        The statistic is also known as the phi coefficient. 
        
        Returns
        -------
        float
            Matthews Correlation Coefficient

        Examples
        --------
        >>> m = model.LogisticRegression()
        >>> m.mathews_corr_coef()
        """

        return sklearn.metrics.matthews_corrcoef(self.y_test, self.y_pred, **kwargs)

    def log_loss(self, **kwargs):
        """
        Log loss, aka logistic loss or cross-entropy loss.

        This is the loss function used in (multinomial) logistic regression and extensions of it
        such as neural networks, defined as the negative log-likelihood of the true labels given a probabilistic classifier’s predictions.
        
        Returns
        -------
        Float
            Log loss

        Examples
        --------
        >>> m = model.LogisticRegression()
        >>> m.log_loss()
        """

        if self.probabilities is not None:
            return sklearn.metrics.log_loss(self.y_test, self.probabilities, **kwargs)
        else:
            return np.nan

    def jaccard(self, **kwargs):
        """
        The Jaccard index, or Jaccard similarity coefficient,
        defined as the size of the intersection divided by the size of the union of two label sets,
        is used to compare set of predicted labels for a sample to the corresponding set of labels in y_true.
        
        Returns
        -------
        float
            Jaccard Score

        Examples
        --------
        >>> m = model.LogisticRegression()
        >>> m.jaccard()
        """

        avg = kwargs.pop("average", "macro")

        if self.multiclass:
            return sklearn.metrics.jaccard_score(
                self.y_test, self.y_pred, average=avg, **kwargs
            )
        else:
            return sklearn.metrics.jaccard_score(self.y_test, self.y_pred, **kwargs)

    def hinge_loss(self, **kwargs):
        """
        Computes the average distance between the model and the data using hinge loss, a one-sided metric that considers only prediction errors.
        
        Returns
        -------
        float
            Hinge loss

        Examples
        --------
        >>> m = model.LogisticRegression()
        >>> m.hinge_loss()
        """

        if hasattr(self.model, "decision_function"):
            return sklearn.metrics.hinge_loss(
                self.y_test, self.model.decision_function(self.x_test), **kwargs
            )
        else:
            return np.nan

    def hamming_loss(self, **kwargs):
        """
        The Hamming loss is the fraction of labels that are incorrectly predicted.
        
        Returns
        -------
        float
            Hamming loss

        Examples
        --------
        >>> m = model.LogisticRegression()
        >>> m.hamming_loss()
        """

        return sklearn.metrics.hamming_loss(self.y_test, self.y_pred, **kwargs)

    def fbeta(self, beta=0.5, **kwargs):
        """
        The F-beta score is the weighted harmonic mean of precision and recall, reaching its optimal value at 1 and its worst value at 0.
        The beta parameter determines the weight of recall in the combined score.
        Beta < 1 lends more weight to precision, while beta > 1 favors recall (beta -> 0 considers only precision, beta -> inf only recall).
        
        Parameters
        ----------
        beta : float, optional
            Weight of precision in harmonic mean, by default 0.5
        
        Returns
        -------
        float
            Fbeta score

        Examples
        --------
        >>> m = model.LogisticRegression()
        >>> m.fbeta()
        """

        avg = kwargs.pop("average", "macro")

        if self.multiclass:
            return sklearn.metrics.fbeta_score(
                self.y_test, self.y_pred, beta, average=avg, **kwargs
            )
        else:
            return sklearn.metrics.fbeta_score(self.y_test, self.y_pred, beta, **kwargs)

    def f1(self, **kwargs):
        """
        The F1 score can be interpreted as a weighted average of the precision and recall, where an F1 score reaches its best value at 1 and worst score at 0. The relative contribution of precision and recall to the F1 score are equal. The formula for the F1 score is:

        F1 = 2 * (precision * recall) / (precision + recall)

        In the multi-class and multi-label case, this is the average of the F1 score of each class with weighting depending on the average parameter.
        
        Returns
        -------
        float
            F1 Score

        Examples
        --------
        >>> m = model.LogisticRegression()
        >>> m.f1()
        """
        avg = kwargs.pop("average", "macro")

        if self.multiclass:
            return sklearn.metrics.f1_score(
                self.y_test, self.y_pred, average=avg, **kwargs
            )
        else:
            return sklearn.metrics.f1_score(self.y_test, self.y_pred, **kwargs)

    def cohen_kappa(self, **kwargs):
        """
        Cohen Kappa tells you how much better is your model over the random classifier that predicts based on class frequencies
        
        This measure is intended to compare labelings by different human annotators, not a classifier versus a ground truth.

        The kappa score (see docstring) is a number between -1 and 1.
        Scores above .8 are generally considered good agreement; zero or lower means no agreement (practically random labels).
        
        Returns
        -------
        float
            Cohen Kappa score.

        Examples
        --------
        >>> m = model.LogisticRegression()
        >>> m.cohen_kappa()
        """

        return sklearn.metrics.cohen_kappa_score(self.y_test, self.y_pred, **kwargs)

    def brier_loss(self, **kwargs):
        """
        Compute the Brier score. The smaller the Brier score, the better, hence the naming with “loss”.  
        Across all items in a set N predictions, the Brier score measures the mean squared difference between (1) the predicted probability assigned to the possible outcomes for item i, and (2) the actual outcome.
        Therefore, the lower the Brier score is for a set of predictions, the better the predictions are calibrated.
        
        The Brier score is appropriate for binary and categorical outcomes that can be structured as true or false,
        but is inappropriate for ordinal variables which can take on three or more values (this is because the Brier score assumes that all possible outcomes are equivalently “distant” from one another)

        Returns
        -------
        float
            Brier loss

        Examples
        --------
        >>> m = model.LogisticRegression()
        >>> m.brier_loss()
        """

        if self.multiclass:
            warnings.warn("Brier Loss can only be used for binary classification.")
            return -999

        return sklearn.metrics.brier_score_loss(self.y_test, self.y_pred, **kwargs)

    def metrics(self, *metrics):
        """
        Measures how well your model performed against certain metrics.

        For multiclassification problems, the 'macro' average is used.

        If a project metrics has been specified, it will display those metrics, otherwise it will display the specified metrics or all metrics.

        For more detailed information and parameters please see the following link: https://scikit-learn.org/stable/modules/classes.html#classification-metrics
        
        Supported metrics are:

            'Accuracy': 'Measures how many observations, both positive and negative, were correctly classified.',
            
            'Balanced Accuracy': 'The balanced accuracy in binary and multiclass classification problems to deal with imbalanced datasets. It is defined as the average of recall obtained on each class.',
            
            'Average Precision': 'Summarizes a precision-recall curve as the weighted mean of precisions achieved at each threshold',
            
            'ROC AUC': 'Shows how good at ranking predictions your model is. It tells you what is the probability that a randomly chosen positive instance is ranked higher than a randomly chosen negative instance.',
            
            'Zero One Loss': 'Fraction of misclassifications.',
            
            'Precision': 'It measures how many observations predicted as positive are positive. Good to use when False Positives are costly.',
            
            'Recall': 'It measures how many observations out of all positive observations have we classified as positive. Good to use when catching call positive occurences, usually at the cost of false positive.',
            
            'Matthews Correlation Coefficient': 'It’s a correlation between predicted classes and ground truth.',
            
            'Log Loss': 'Difference between ground truth and predicted score for every observation and average those errors over all observations.',
            
            'Jaccard': 'Defined as the size of the intersection divided by the size of the union of two label sets, is used to compare set of predicted labels for a sample to the corresponding set of true labels.',
            
            'Hinge Loss': 'Computes the average distance between the model and the data using hinge loss, a one-sided metric that considers only prediction errors.',
            
            'Hamming Loss': 'The Hamming loss is the fraction of labels that are incorrectly predicted.',
            
            'F-Beta': 'It’s the harmonic mean between precision and recall, with an emphasis on one or the other. Takes into account both metrics, good for imbalanced problems (spam, fraud, etc.).',
            
            'F1': 'It’s the harmonic mean between precision and recall. Takes into account both metrics, good for imbalanced problems (spam, fraud, etc.).',
            
            'Cohen Kappa': 'Cohen Kappa tells you how much better is your model over the random classifier that predicts based on class frequencies. Works well for imbalanced problems.',
            
            'Brier Loss': 'It is a measure of how far your predictions lie from the true values. Basically, it is a mean square error in the probability space.'
        
        Parameters
        ----------
        metrics : str(s), optional
            Specific type of metrics to view

        Examples
        --------
        >>> m = model.LogisticRegression()
        >>> m.metrics()
        >>> m.metrics('F1', 'F-Beta')
        """

        from aethos.modelling.constants import CLASS_METRICS_DESC

        metric_list = {
            "Accuracy": self.accuracy(),
            "Balanced Accuracy": self.balanced_accuracy(),
            "Average Precision": self.average_precision(),
            "ROC AUC": self.roc_auc(),
            "Zero One Loss": self.zero_one_loss(),
            "Precision": self.precision(),
            "Recall": self.recall(),
            "Matthews Correlation Coefficient": self.matthews_corr_coef(),
            "Log Loss": self.log_loss(),
            "Jaccard": self.jaccard(),
            "Hinge Loss": self.hinge_loss(),
            "Hamming Loss": self.hamming_loss(),
            "F-Beta": self.fbeta(),
            "F1": self.f1(),
            "Cohen Kappa": self.cohen_kappa(),
            "Brier Loss": self.brier_loss(),
        }

        metric_table = pd.DataFrame(
            index=metric_list.keys(),
            columns=[self.model_name],
            data=metric_list.values(),
        )
        metric_table["Description"] = [
            CLASS_METRICS_DESC[x] for x in metric_table.index
        ]

        pd.set_option("display.max_colwidth", -1)

        if not metrics and _global_config["project_metrics"]:  # pragma: no cover
            filt_metrics = _global_config["project_metrics"]
        else:
            filt_metrics = list(metrics) if metrics else metric_table.index

        if self.report:
            self.report.write_metrics(metric_table.loc[filt_metrics, :].round(3))

        return metric_table.loc[filt_metrics, :].round(3)

    def confusion_matrix(
        self,
        title=None,
        normalize=False,
        hide_counts=False,
        x_tick_rotation=0,
        figsize=None,
        cmap="Blues",
        title_fontsize="large",
        text_fontsize="medium",
        output_file="",
    ):
        """
        Prints a confusion matrix as a heatmap.
    
        Arguments
        ---------
        title : str
            The text to display at the top of the matrix, by default 'Confusion Matrix'

        normalize : bool
            If False, plot the raw numbers
            If True, plot the proportions,
            by default False

        hide_counts : bool
            If False, display the counts and percentage
            If True, hide display of the counts and percentage
            by default, False

        x_tick_rotation : int
            Degree of rotation to rotate the x ticks
            by default, 0

        figsize : tuple(int, int)
            Size of the figure
            by default, None

        cmap : str   
            The gradient of the values displayed from matplotlib.pyplot.cm
            see http://matplotlib.org/examples/color/colormaps_reference.html
            plt.get_cmap('jet') or plt.cm.Blues
            by default, 'Blues'

        title_fontsize : str
            Size of the title, by default 'large'

        text_fontsize : str
            Size of the text of the rest of the plot, by default 'medium' 

        output_file: str
            Output file name including extension (.png, .jpg, etc.) to save image as.

        Examples
        --------
        >>> m = model.LogisticRegression()
        >>> m.confusion_matrix()      
        >>> m.confusion_matrix(normalize=True)      
        """

        import seaborn as sns

        y_true = self.y_test
        y_pred = self.y_pred

        if figsize:
            plt.figure(figsize=figsize)

        confusion_matrix = sklearn.metrics.confusion_matrix(y_true, y_pred)

        if normalize:
            confusion_matrix = (
                confusion_matrix.astype("float")
                / confusion_matrix.sum(axis=1)[:, np.newaxis]
            )

        accuracy = np.trace(confusion_matrix) / float(np.sum(confusion_matrix))
        mis_class = 1 - accuracy

        if title:
            plt.title(title, fontsize=title_fontsize)
        elif normalize:
            plt.title("Normalized Confusion Matrix", fontsize=title_fontsize)
        else:
            plt.title("Confusion Matrix", fontsize=title_fontsize)

        cm_sum = np.sum(confusion_matrix, axis=1)
        cm_perc = confusion_matrix / cm_sum.astype(float) * 100
        nrows, ncols = confusion_matrix.shape

        if not hide_counts:
            annot = np.zeros_like(confusion_matrix).astype("str")

            for i in range(nrows):
                for j in range(ncols):
                    c = confusion_matrix[i, j]
                    p = cm_perc[i, j]
                    if i == j:
                        s = cm_sum[i]
                        annot[i, j] = "{:.2f}%\n{}/{}".format(float(p), int(c), int(s))
                    elif c == 0:
                        annot[i, j] = ""
                    else:
                        annot[i, j] = "{:.2f}%\n{}".format(p, c)
        else:
            annot = np.zeros_like(confusion_matrix, dtype=str)

        df_cm = pd.DataFrame(confusion_matrix, index=self.classes, columns=self.classes)

        heatmap = sns.heatmap(
            df_cm, annot=annot, square=True, cmap=plt.cm.get_cmap(cmap), fmt=""
        )

        plt.tight_layout()
        plt.ylabel("True label", fontsize=text_fontsize)
        plt.xlabel(
            "Predicted label\naccuracy={:0.4f}; misclassified={:0.4f}".format(
                accuracy, mis_class
            ),
            fontsize=text_fontsize,
        )
        plt.xticks(
            np.arange(len(self.classes)) + 0.5, self.classes, rotation=x_tick_rotation
        )
        plt.show()

        if output_file or _global_config["track_experiments"]:  # pragma: no cover
            heatmap.figure.savefig(
                os.path.join(IMAGE_DIR, self.model_name, output_file)
            )

            if self.report:
                self.report.write_image(
                    os.path.join(IMAGE_DIR, self.model_name, output_file)
                )

        if _global_config["track_experiments"]:  # pragma: no cover
            track_artifacts(self.run_id, self.model_name)

    def roc_curve(self, title=True, output_file=""):
        """
        Plots an ROC curve and displays the ROC statistics (area under the curve).

        Parameters
        ----------
        figsize : tuple(int, int), optional
            Figure size, by default (600,450)

        title : bool
            Whether to display title, by default True

        output_file : str, optional
            If a name is provided save the plot to an html file, by default ''

        Examples
        --------
        >>> m = model.LogisticRegression()
        >>> m.roc_curve()
        """

        if self.multiclass:
            raise NotImplementedError(
                "ROC Curve not implemented for multiclassification problems yet."
            )
        else:
            roc_auc = self.roc_auc()

            roc_plot = sklearn.metrics.plot_roc_curve(
                self.model, self.x_test, self.y_test
            )
            roc_plot.ax_.set_xlabel("False Positive Rate or (1 - Specifity)")
            roc_plot.ax_.set_ylabel("True Positive Rate or (Sensitivity)")
            if title:
                roc_plot.figure_.suptitle("ROC Curve (area = {:.2f})".format(roc_auc))

        if output_file:  # pragma: no cover
            roc_plot.figure_.savefig(
                os.path.join(IMAGE_DIR, self.model_name, output_file)
            )

            if self.report:
                self.report.write_image(
                    os.path.join(IMAGE_DIR, self.model_name, output_file)
                )

        if _global_config["track_experiments"]:  # pragma: no cover
            track_artifacts(self.run_id, self.model_name)

        return roc_plot

    def classification_report(self):
        """
        Prints and logs the classification report.

        The classification report displays and logs the information in this format:

                    precision    recall  f1-score   support

                    1       1.00      0.67      0.80         3
                    2       0.00      0.00      0.00         0
                    3       0.00      0.00      0.00         0

            micro avg       1.00      0.67      0.80         3
            macro avg       0.33      0.22      0.27         3
         weighted avg       1.00      0.67      0.80         3

        Examples
        --------
        >>> m = model.LogisticRegression()
        >>> m.classification_report()
        """

        classification_report = sklearn.metrics.classification_report(
            self.y_test, self.y_pred, target_names=self.classes, digits=2
        )

        if self.report:
            self.report.report_classification_report(classification_report)

        print(classification_report)


class RegressionModel(ModelBase):
    def __init__(
        self,
        model_object,
        model_name,
        model,
        predictions_col,
        shap_values=None,
        pool=None,
        run_id=None,
    ):

        self.y_train = model_object.y_train
        self.y_test = (
            model_object.y_test
            if model_object.x_test is not None
            else model_object.y_train
        )

        super().__init__(model_object, model, model_name, shap_values=shap_values)

        self.y_pred = (
            model_object.x_train_results[predictions_col]
            if self.x_test is None
            else model_object.x_test_results[predictions_col]
        )

        if self.report:
            self.report.write_header(f"Analyzing Model {self.model_name.upper()}: ")

        self.features = self.x_test.columns
        self.pool = pool
        self.run_id = run_id

    def explained_variance(self, multioutput="uniform_average", **kwargs):
        """
        Explained variance regression score function

        Best possible score is 1.0, lower values are worse.
        
        Parameters
        ----------
        multioutput : string in [‘raw_values’, ‘uniform_average’, ‘variance_weighted’] or array-like of shape (n_outputs)
            Defines aggregating of multiple output scores. Array-like value defines weights used to average scores.

            ‘raw_values’ :
                Returns a full set of scores in case of multioutput input.

            ‘uniform_average’ :
                Scores of all outputs are averaged with uniform weight.

            ‘variance_weighted’ :
                Scores of all outputs are averaged, weighted by the variances of each individual output.

            By default 'uniform_average'
        
        Returns
        -------
        float
            Explained Variance

        Examples
        --------
        >>> m = model.LinearRegression()
        >>> m.explained_variance()
        """

        return sklearn.metrics.explained_variance_score(
            self.y_test, self.y_pred, multioutput="uniform_average", **kwargs
        )

    def max_error(self):
        """
        Returns the single most maximum residual error.
        
        Returns
        -------
        float
            Max error

        Examples
        --------
        >>> m = model.LinearRegression()
        >>> m.max_error()
        """

        return sklearn.metrics.max_error(self.y_test, self.y_pred)

    def mean_abs_error(self, **kwargs):
        """
        Mean absolute error.
        
        Returns
        -------
        float
            Mean absolute error.

        Examples
        --------
        >>> m = model.LinearRegression()
        >>> m.mean_abs_error()
        """

        return sklearn.metrics.mean_absolute_error(self.y_test, self.y_pred)

    def mean_sq_error(self, **kwargs):
        """
        Mean squared error.
        
        Returns
        -------
        float
            Mean squared error.

        Examples
        --------
        >>> m = model.LinearRegression()
        >>> m.mean_sq_error()
        """

        return sklearn.metrics.mean_squared_error(self.y_test, self.y_pred)

    def mean_sq_log_error(self, **kwargs):
        """
        Mean squared log error.
        
        Returns
        -------
        float
            Mean squared log error.

        Examples
        --------
        >>> m = model.LinearRegression()
        >>> m.mean_sq_log_error()
        """

        try:
            return sklearn.metrics.mean_squared_log_error(self.y_test, self.y_pred)
        except ValueError as e:
            warnings.warn(
                "Mean Squared Logarithmic Error cannot be used when targets contain negative values."
            )
            return -999

    def median_abs_error(self, **kwargs):
        """
        Median absolute error.
        
        Returns
        -------
        float
            Median absolute error.

        Examples
        --------
        >>> m = model.LinearRegression()
        >>> m.median_abs_error()
        """

        return sklearn.metrics.median_absolute_error(self.y_test, self.y_pred)

    def r2(self, **kwargs):
        """
        R^2 (coefficient of determination) regression score function.

        R-squared (R2) is a statistical measure that represents the proportion of the variance for a dependent variable
        that is explained by an independent variable or variables in a regression model.

        Best possible score is 1.0 and it can be negative (because the model can be arbitrarily worse).
        A constant model that always predicts the expected value of y, disregarding the input features, would get a R^2 score of 0.0.
        
        Returns
        -------
        float
            R2 coefficient.

        Examples
        --------
        >>> m = model.LinearRegression()
        >>> m.r2()
        """

        return sklearn.metrics.r2_score(self.y_test, self.y_pred)

    def smape(self, **kwargs):
        """
        Symmetric mean absolute percentage error.

        It is an accuracy measure based on percentage (or relative) errors.
        
        Returns
        -------
        float
            SMAPE

        Examples
        --------
        >>> m = model.LinearRegression()
        >>> m.smape()
        """

        return (
            1
            / len(self.y_test)
            * np.sum(
                2
                * np.abs(self.y_pred - self.y_test)
                / (np.abs(self.y_test) + np.abs(self.y_pred))
            )
        )

    def root_mean_sq_error(self):
        """
        Root mean squared error.

        Calculated by taking the square root of the Mean Squared Error.

        Returns
        -------
        float
            Root mean squared error.

        Examples
        --------
        >>> m = model.LinearRegression()
        >>> m.root_mean_sq_error()
        """

        return math.sqrt(self.mean_sq_error())

    def metrics(self, *metrics):
        """
        Measures how well your model performed against certain metrics.

        If a project metrics has been specified, it will display those metrics, otherwise it will display the specified metrics or all metrics.

        For more detailed information and parameters please see the following link: https://scikit-learn.org/stable/modules/classes.html#regression-metrics
        
        Supported metrics are:
            'Explained Variance': 'Explained variance regression score function. Best possible score is 1.0, lower values are worse.',
            
            'Max Error': 'Returns the single most maximum residual error.',
            
            'Mean Absolute Error': 'Postive mean value of all residuals',
            
            'Mean Squared Error': 'Mean of the squared sum the residuals',
            
            'Root Mean Sqaured Error': 'Square root of the Mean Squared Error',
            
            'Mean Squared Log Error': 'Mean of the squared sum of the log of all residuals',
            
            'Median Absolute Error': 'Postive median value of all residuals',
            
            'R2': 'R-squared (R2) is a statistical measure that represents the proportion of the variance for a dependent variable that is explained by an independent variable or variables in a regression model.',
            
            'SMAPE': 'Symmetric mean absolute percentage error. It is an accuracy measure based on percentage (or relative) errors.'

        Parameters
        ----------
        metrics : str(s), optional
            Specific type of metrics to view

        Examples
        --------
        >>> m = model.LinearRegression()
        >>> m.metrics()
        >>> m.metrics('SMAPE', 'Root Mean Squared Error')
        """

        from aethos.modelling.constants import REG_METRICS_DESC

        metric_list = {
            "Explained Variance": self.explained_variance(),
            "Max Error": self.max_error(),
            "Mean Absolute Error": self.mean_abs_error(),
            "Mean Squared Error": self.mean_sq_error(),
            "Root Mean Sqaured Error": self.root_mean_sq_error(),
            "Mean Squared Log Error": self.mean_sq_log_error(),
            "Median Absolute Error": self.median_abs_error(),
            "R2": self.r2(),
            "SMAPE": self.smape(),
        }

        metric_table = pd.DataFrame(
            index=metric_list.keys(),
            columns=[self.model_name],
            data=metric_list.values(),
        )
        metric_table["Description"] = [REG_METRICS_DESC[x] for x in metric_table.index]

        pd.set_option("display.max_colwidth", -1)

        if not metrics and _global_config["project_metrics"]:  # pragma: no cover
            filt_metrics = _global_config["project_metrics"]
        else:
            filt_metrics = list(metrics) if metrics else metric_table.index

        if self.report:
            self.report.write_metrics(metric_table.loc[filt_metrics, :].round(3))

        return metric_table.loc[filt_metrics, :].round(3)



import warnings
import os
import shutil

warnings.simplefilter("ignore", FutureWarning)
warnings.simplefilter("ignore", DeprecationWarning)

aethos_home = os.path.join(os.path.expanduser("~"), ".aethos")
config_home = os.path.join(aethos_home, "config.yml")
pkg_directory = os.path.dirname(__file__)

# Create the config file
if not os.path.exists(config_home):
    os.makedirs(aethos_home)
    shutil.copyfile(
        os.path.join(pkg_directory, "config", "config.yml"), os.path.join(config_home)
    )

import pandas as pd
from IPython import get_ipython
import plotly.io as pio

# let init-time option registration happen
import aethos.config.config_init
from aethos.config.config import (
    describe_option,
    get_option,
    options,
    reset_option,
    set_option,
)


from .core import Data
from .modelling import Model

pd.options.mode.chained_assignment = None
pio.templates.default = "plotly_white"

__all__ = ["Data", "Model"]

shell = get_ipython().__class__.__name__

if shell == "ZMQInteractiveShell":
    import shap

    shap.initjs()



import os

import yaml
from IPython import get_ipython

from aethos.util import _make_dir

pkg_directory = os.path.dirname(__file__)

with open(
    os.path.join(os.path.expanduser("~"), ".aethos", "config.yml"), "r"
) as ymlfile:
    cfg = yaml.safe_load(ymlfile)

with open(f"{pkg_directory}/technique_reasons.yml", "r") as stream:
    technique_reason_repo = yaml.safe_load(stream)

shell = get_ipython().__class__.__name__

DEFAULT_MODEL_DIR = os.path.join(os.path.expanduser("~"), ".aethos", "models")
DEFAULT_REPORT_DIR = os.path.join(os.path.expanduser("~"), ".aethos", "reports")
DEFAULT_IMAGE_DIR = os.path.join(os.path.expanduser("~"), ".aethos", "images")
DEFAULT_EXPERIMENTS_DIR = "file:" + os.path.join(
    os.path.expanduser("~"), ".aethos", "experiments", "mlruns"
)
DEFAULT_DEPLOYMENTS_DIR = os.path.join(os.path.expanduser("~"), ".aethos", "projects")


def _make_image_dir():

    if not cfg["images"]["dir"]:
        image_dir = DEFAULT_IMAGE_DIR
    else:
        image_dir = cfg["images"]["dir"]

    _make_dir(image_dir)

    return image_dir


IMAGE_DIR = _make_image_dir()


def _make_experiment_dir():  # pragma: no cover

    if not cfg["mlflow"]["dir"]:
        exp_dir = DEFAULT_EXPERIMENTS_DIR
    else:
        exp_dir = cfg["mlflow"]["dir"]

    if exp_dir.startswith("file:"):
        _make_dir(exp_dir[5:])

    return exp_dir


EXP_DIR = _make_experiment_dir()



import os

from aethos.config import DEFAULT_DEPLOYMENTS_DIR, cfg
from aethos.util import _make_dir


def _create_dir():
    """
    Creates the projects directory.
    
    Parameters
    ----------
    project_dir : str
        Full path of the project dir.

    name : str
        Name of the project
    """

    if not cfg["models"]["deployment_dir"]:
        dep_dir = DEFAULT_DEPLOYMENTS_DIR
    else:
        dep_dir = cfg["models"]["deployment_dir"]

    _make_dir(dep_dir)

    return dep_dir

def _create_project_dir(project_dir: str, name: str):
    """
    Creates the projects directory.
    
    Parameters
    ----------
    project_dir : str
        Full path of the project dir.

    name : str
        Name of the project
    """

    project_dir = os.path.join(project_dir, name)

    _make_dir(project_dir)

    return project_dir

def _get_model_type_kwarg(model):
    """
    Helper function to determine what model to import in requirements.txt service file.
    """

    import xgboost as xgb
    import catboost as cb
    import lightgbm as lgb

    kwargs = {
        'xgboost': False,
        'catboost': False,
        'lgbm': False,
    }

    if isinstance(model, xgb.XGBModel):
        kwargs['xgboost'] = True
    
    if isinstance(model, cb.CatBoost):
        kwargs['catboost'] = True

    if isinstance(model, lgb.LGBMModel):
        kwargs['lgbm'] = True

    return kwargs



import os
from pathlib import Path

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import plotly.express as px
import seaborn as sns
from aethos.config import IMAGE_DIR, cfg
from aethos.util import _make_dir


def raincloud(col: str, target_col: str, data: pd.DataFrame, output_file="", **params):
    """
    Visualizes 2 columns using raincloud.
    
    Parameters
    ----------
    col : str
        Column name of general data

    target_col : str
        Column name of measurable data, numerical

    data : Dataframe
        Dataframe of the data

    params: dict
        Parameters for the RainCloud visualization

    ouput_file : str
        Output file name for the image including extension (.jpg, .png, etc.)
    """

    import ptitprince as pt

    fig, ax = plt.subplots(figsize=(12, 8))

    if not params:
        params = {
            "pointplot": True,
            "width_viol": 0.8,
            "width_box": 0.4,
            "orient": "h",
            "move": 0.0,
            "ax": ax,
        }

    ax = pt.RainCloud(x=col, y=target_col, data=data.infer_objects(), **params)

    if output_file:  # pragma: no cover
        fig.savefig(os.path.join(IMAGE_DIR, output_file))

    return ax


def barplot(
    x,
    y,
    data,
    method=None,
    orient="v",
    barmode="relative",
    title="",
    yaxis_params=None,
    xaxis_params=None,
    output_file="",
    **barplot_kwargs,
):
    """
    Visualizes a bar plot.
    
    Parameters
    ----------
    x : str
        Column name for the x axis.

    y : str, list
        Columns for the y axis

    data : Dataframe
        Dataset

    method : str
        Method to aggregate groupy data
        Examples: min, max, mean, etc., optional
        by default None

    orient : str, optional
        Orientation of graph, 'h' for horizontal
        'v' for vertical, by default 'v'

    barmode : str {'relative', 'overlay', 'group', 'stack'}
        Relative is a normal barplot
        Overlay barplot shows positive values above 0 and negative values below 0
        Group are the bars beside each other.
        Stack groups the bars on top of each other
        by default 'relative'

    yaxis_params : dict
        Parameters for the y axis

    xaxis_params : dict
        Parameters for the x axis
    """

    import plotly.graph_objects as go

    if isinstance(y, str):
        y = [y]

    data_copy = data.copy()

    if method:
        data_copy = data_copy.groupby(x, as_index=False)
        data_copy = getattr(data_copy, method)()

    fig = go.Figure()

    if not xaxis_params:
        xaxis_params = {"title": x.capitalize()}

    if not yaxis_params:
        if method:
            yaxis_params = {"title": method.capitalize()}

    if len(y) > 1:

        for col in y:
            fig.add_trace(
                go.Bar(
                    x=data_copy[x],
                    y=data[col],
                    name=col,
                    width=[0.1] * len(data_copy[x]),
                    orientation=orient,
                )
            )

        fig.update_layout(
            title=title,
            xaxis=xaxis_params,
            yaxis=yaxis_params,
            barmode=barmode,
            bargap=0.8,  # gap between bars of adjacent location coordinates.
            bargroupgap=0.1,  # gap between bars of the same location coordinate
        )

    else:
        for col in y:
            fig.add_trace(
                go.Bar(
                    x=data_copy[x],
                    y=data[col],
                    name=col,
                    width=[0.5] * len(data_copy[x]),
                    orientation=orient,
                )
            )

        fig.update_layout(
            title=title, xaxis=xaxis_params, yaxis=yaxis_params, barmode=barmode,
        )

    if output_file:  # pragma: no cover
        fig.write_image(os.path.join(IMAGE_DIR, output_file))

    return fig


def scatterplot(
    x: str,
    y: str,
    z=None,
    data=None,
    category=None,
    title="Scatter Plot",
    output_file="",
    **scatterplot_kwargs,
):
    """
    Plots a scatter plot.
    
    Parameters
    ----------
    x : str
        X axis column

    y : str
        Y axis column

    z: str
        Z axis column

    data : Dataframe
        Dataframe

    category : str, optional
        Category to group your data, by default None

    title : str, optional
        Title of the plot, by default 'Scatterplot'

    size : int or str, optional
        Size of the circle, can either be a number
        or a column name to scale the size, by default 8

    output_file : str, optional
        If a name is provided save the plot to an html file, by default ''
    """

    if z is None:
        fig = px.scatter(
            data, x=x, y=y, color=category, title=title, **scatterplot_kwargs
        )

    else:
        fig = px.scatter_3d(
            data, x=x, y=y, z=z, color=category, title=title, **scatterplot_kwargs
        )

    if output_file:  # pragma: no cover
        fig.write_image(os.path.join(IMAGE_DIR, output_file))

    return fig


def lineplot(
    x: str, y: str, z: str, data, category=None, title="Line Plot", output_file="", **lineplot_kwargs
):
    """
    Plots a line plot.
    
    Parameters
    ----------
    x : str
        X axis column

    y : str
        Y axis column

    z : str
        Z axis column

    data : Dataframe
        Dataframe

    category : str
        Category column to draw multiple line plots of

    title : str, optional
        Title of the plot, by default 'Line Plot'

    output_file : str, optional
        If a name is provided save the plot to an html file, by default ''
    """

    if z is None:
        fig = px.line(
            data, x=x, y=y, color=category, title=title, **lineplot_kwargs
        )

        fig.data[0].update(mode='markers+lines')

    else:
        fig = px.line_3d(
            data, x=x, y=y, z=z, color=category, title=title, **lineplot_kwargs
        )

    if output_file:  # pragma: no cover
        fig.write_image(os.path.join(IMAGE_DIR, output_file))

    return fig

def correlation_matrix(
    df, data_labels=False, hide_mirror=False, output_file="", **kwargs
):
    """
    Plots a correlation matrix.
    
    Parameters
    ----------
    df : DataFrame
        Data

    data_labels : bool, optional
        Whether to display the correlation values, by default False

    hide_mirror : bool, optional
        Whether to display the mirroring half of the correlation plot, by default False

    ouput_file : str
        Output file name for the image including extension (.jpg, .png, etc.)
    """

    corr = df.corr()

    fig, ax = plt.subplots(figsize=(11, 9))

    if hide_mirror:
        mask = np.zeros_like(corr, dtype=np.bool)
        mask[np.triu_indices_from(mask)] = True
    else:
        mask = None

    cmap = sns.diverging_palette(220, 10, as_cmap=True)

    sns.heatmap(
        corr,
        cmap=cmap,
        vmax=0.3,
        center=0,
        square=True,
        mask=mask,
        annot=data_labels,
        fmt=".2f",
        linewidths=0.5,
        cbar_kws={"shrink": 0.5},
        **kwargs,
    )

    if output_file:  # pragma: no cover
        fig.savefig(os.path.join(IMAGE_DIR, output_file))

    return ax

def pairplot(
    df,
    kind="scatter",
    diag_kind="auto",
    upper_kind=None,
    lower_kind=None,
    hue=None,
    output_file=None,
    **kwargs,
):
    """
    Plots pairplots of the variables in the DataFrame
    
    Parameters
    ----------
    df : DataFrame
        Data

    kind : {'scatter', 'reg'}, optional
        Type of plot for off-diag plots, by default 'scatter'

    diag_kind : {'auto', 'hist', 'kde'}, optional
        Type of plot for diagonal, by default 'auto'

    upper_kind : str {'scatter', 'kde'}, optional
        Type of plot for upper triangle of pair plot, by default None

    lower_kind : str {'scatter', 'kde'}, optional
        Type of plot for lower triangle of pair plot, by default None

    hue : str, optional
        Column to colour points by, by default None

    ouput_file : str
        Output file name for the image including extension (.jpg, .png, etc.)
    """

    plot_mapping = {"scatter": plt.scatter, "kde": sns.kdeplot, "hist": sns.distplot}

    palette = kwargs.pop("color", sns.color_palette("pastel"))

    if upper_kind or lower_kind:
        assert upper_kind is not None, "upper_kind cannot be None."
        assert lower_kind is not None, "lower_kind cannot be None."
        assert diag_kind is not "auto", "diag_kind must be either `hist` or `kde`."

        g = sns.PairGrid(df, hue=hue, palette=palette, **kwargs)
        g = g.map_upper(plot_mapping[upper_kind])
        g = g.map_diag(plot_mapping[diag_kind])
        g = g.map_lower(plot_mapping[lower_kind])
        g = g.add_legend()

    else:
        g = sns.pairplot(
            df, kind=kind, diag_kind=diag_kind, hue=hue, palette=palette, **kwargs
        )

    if output_file:  # pragma: no cover
        g.savefig(os.path.join(IMAGE_DIR, output_file))

    return g


def jointplot(x, y, df, kind="scatter", output_file="", **kwargs):
    """
    Plots a joint plot of 2 variables.
    
    Parameters
    ----------
    x : str
        X axis column

    y : str
        y axis column

    df : DataFrame
        Data

    kind : { “scatter” | “reg” | “resid” | “kde” | “hex” }, optional
        Kind of plot to draw, by default 'scatter'

    ouput_file : str
        Output file name for the image including extension (.jpg, .png, etc.)
    """

    # NOTE: Ignore the deprecation warning for showing the R^2 statistic until Seaborn reimplements it
    import warnings
    from scipy import stats

    warnings.simplefilter("ignore", UserWarning)

    sns.set(style="ticks", color_codes=True)
    color = kwargs.pop("color", "crimson")

    g = sns.jointplot(x=x, y=y, data=df, kind=kind, color=color, **kwargs).annotate(
        stats.pearsonr
    )

    if output_file:  # pragma: no cover
        g.savefig(os.path.join(IMAGE_DIR, output_file))

    return g


def histogram(x: list, x_train: pd.DataFrame, x_test=None, output_file="", **kwargs):
    """
    Plots a histogram.
    
    Parameters
    ----------
    x : list
        Columns to plot histogram for.

    x_train : pd.DataFrame
        Dataframe of the data.

    x_test : pd.DataFrame
        Dataframe of the data.

    ouput_file : str
        Output file name for the image including extension (.jpg, .png, etc.)
    """

    import math

    sns.set(style="ticks", color_codes=True)
    sns.set_palette(sns.color_palette("pastel"))

    # Make the single plot look pretty
    if len(x) == 1:
        data = x_train[~x_train[x[0]].isnull()]
        g = sns.distplot(data[x], label="Train Data", **kwargs)

        if x_test is not None:
            data = x_test[~x_test[x[0]].isnull()]
            g = sns.distplot(data[x], label="Test Data", **kwargs)
            g.legend(loc="upper right")

        g.set_title(f"Histogram for {x[0].capitalize()}")

    else:
        n_cols = 2
        n_rows = math.ceil(len(x) / n_cols)

        _, ax = plt.subplots(n_rows, n_cols, figsize=(30, 5 * n_cols))

        for ax, col in zip(ax.flat, x):
            g = None

            data = x_train[~x_train[col].isnull()]
            g = sns.distplot(data[col], ax=ax, label="Train Data", **kwargs)

            if x_test is not None:
                data = x_test[~x_test[col].isnull()]
                g = sns.distplot(data[col], ax=ax, label="Test Data", **kwargs)
                ax.legend(loc="upper right")

            ax.set_title(f"Histogram for {col.capitalize()}", fontsize=20)

        plt.tight_layout()

    if output_file:  # pragma: no cover
        g.figure.savefig(os.path.join(IMAGE_DIR, output_file))

    return g


def viz_clusters(
    data: pd.DataFrame, algo: str, category: str, dim=2, output_file="", **kwargs
):
    """
    Visualize clusters
    
    Parameters
    ----------
    data : pd.DataFrame
        Data

    algo : str {'tsne', 'lle', 'pca', 'tsvd'}, optional
        Algorithm to reduce the dimensions by, by default 'tsne'

    category : str
        Column name of the labels/data points to highlight in the plot

    dim : int {2, 3}
        Dimensions of the plot to show, either 2d or 3d, by default 2

    output_file : str, optional
        Output file name for image with extension (i.e. jpeg, png, etc.)
    """

    from sklearn.manifold import LocallyLinearEmbedding, TSNE
    from sklearn.decomposition import PCA, TruncatedSVD

    if dim != 2 and dim != 3:
        raise ValueError("Dimension must be either 2d (2) or 3d (3)")

    algorithms = {
        "tsne": TSNE(n_components=dim, random_state=42,),
        "lle": LocallyLinearEmbedding(n_components=dim, random_state=42,),
        "pca": PCA(n_components=dim, random_state=42,),
        "tsvd": TruncatedSVD(n_components=dim, random_state=42,),
    }

    reducer = algorithms[algo]
    reduced_df = pd.DataFrame(reducer.fit_transform(data.drop(category, axis=1)))
    reduced_df.columns = map(str, reduced_df.columns)
    reduced_df[category] = data[category]
    reduced_df[category] = reduced_df[category].astype(str)

    if dim == 2:
        fig = scatterplot(
            "0",
            "1",
            data=reduced_df,
            category=category,
            output_file=output_file,
            **kwargs,
        )
    else:
        fig = scatterplot(
            "0",
            "1",
            "2",
            data=reduced_df,
            category=category,
            output_file=output_file,
            **kwargs,
        )

    return fig


def boxplot(
    x=None, y=None, data=None, orient="v", title="", output_file="", **boxplot_kwargs,
):
    """
    Plots a box plot

    Parameters
    ----------
    x : str, optional
        Column from data, by default None

    y : str, optional
        Column from data, by default None

    data : DataFrame, optional
        Data, by default None

    orient : str, optional
        Orientation of the plot, by default "v"

    title : str, optional
        Title of the plot, by default ''

    output_file : str, optional
        File name, by default ""
    """

    fig = px.box(data, x=x, y=y, orientation=orient, title=title, **boxplot_kwargs)

    if output_file:  # pragma: no cover
        fig.write_image(os.path.join(IMAGE_DIR, output_file))

    return fig

def violinplot(
    x=None, y=None, data=None, orient="v", title="", output_file="", **violin_kwargs,
):
    """
    Plots a violin plot.

    Parameters
    ----------
    x : str, optional
        Column from data, by default None

    y : str, optional
        Column from data, by default None

    data : DataFrame, optional
        Data, by default None

    orient : str, optional
        Orientation of the plot, by default "v"

    title : str, optional
        Title of the plot, by default ''

    output_file : str, optional
        File name, by default ""
    """

    fig = px.violin(data, x=x, y=y, orientation=orient, title=title, **violin_kwargs)

    if output_file:  # pragma: no cover
        fig.write_image(os.path.join(IMAGE_DIR, output_file))

    return fig


def create_table(matrix, index, output_file, **kwargs):
    """
    Creates a table using plotly.
    
    Parameters
    ----------
    matrix : 2d array
        Table values
    """

    from plotly.tools import FigureFactory as FF

    table = FF.create_table(matrix, index=True)

    if output_file:  # pragma: no cover
        table.write_image(os.path.join(IMAGE_DIR, output_file))

    table.show()



from aethos.config import technique_reason_repo
from aethos.feature_engineering import categorical as cat
from aethos.feature_engineering import numeric as num
from aethos.feature_engineering import text
from aethos.feature_engineering import util
from aethos.util import _input_columns, label_encoder


class Feature(object):
    def onehot_encode(
        self, *list_args, list_of_cols=[], keep_col=True, **onehot_kwargs
    ):
        """
        Creates a matrix of converted categorical columns into binary columns of ones and zeros.

        For more info see: https://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.OneHotEncoder.html

        If a list of columns is provided use the list, otherwise use arguments.
    
        Parameters
        ----------
        list_args : str(s), optional
            Specific columns to apply this technique to.

        list_of_cols : list, optional
            A list of specific columns to apply this technique to., by default []

        keep_col : bool
            A parameter to specify whether to drop the column being transformed, by default
            keep the column, True

        categories : ‘auto’ or a list of array-like, default=’auto’
            Categories (unique values) per feature:

                ‘auto’ : Determine categories automatically from the training data.

                list : categories[i] holds the categories expected in the ith column. The passed categories should not mix strings and numeric values within a single feature, and should be sorted in case of numeric values.

            The used categories can be found in the categories_ attribute.

        drop : ‘first’ or a array-like of shape (n_features,), default=None
            Specifies a methodology to use to drop one of the categories per feature. This is useful in situations where perfectly collinear features cause problems, such as when feeding the resulting data into a neural network or an unregularized regression.

                None : retain all features (the default).

                ‘first’ : drop the first category in each feature. If only one category is present, the feature will be dropped entirely.

                array : drop[i] is the category in feature X[:, i] that should be dropped.

        sparsebool : default=True
            Will return sparse matrix if set True else will return an array.

        dtype : number type, default=np.float
            Desired dtype of output.

        handle_unknown: {‘error’, ‘ignore’}, default='ignore'
            Whether to raise an error or ignore if an unknown categorical feature is present during transform (default is to raise).
            When this parameter is set to ‘ignore’ and an unknown category is encountered during transform, the resulting one-hot encoded columns for this feature will be all zeros.
            In the inverse transform, an unknown category will be denoted as None.

        Returns
        -------
        Data:
            Returns a deep copy of the Data object.

        Examples
        --------
        >>> data.onehot_encode('col1', 'col2', 'col3')
        >>> data.onehot_encode('col1', 'col2', 'col3', drop='first')
        """

        report_info = technique_reason_repo["feature"]["categorical"]["onehotencode"]

        ## If a list of columns is provided use the list, otherwise use arguemnts.
        list_of_cols = _input_columns(list_args, list_of_cols)

        (self.x_train, self.x_test,) = cat.feature_one_hot_encode(
            x_train=self.x_train,
            x_test=self.x_test,
            list_of_cols=list_of_cols,
            keep_col=keep_col,
            **onehot_kwargs,
        )
        if self.report is not None:
            self.report.report_technique(report_info, list_of_cols)

        return self.copy()

    def tfidf(self, *list_args, list_of_cols=[], keep_col=True, **tfidf_kwargs):
        """
        Creates a matrix of the tf-idf score for every word in the corpus as it pertains to each document.

        The higher the score the more important a word is to a document, the lower the score (relative to the other scores)
        the less important a word is to a document.

        For more information see: https://scikit-learn.org/stable/modules/generated/sklearn.feature_extraction.text.TfidfVectorizer.html

        If a list of columns is provided use the list, otherwise use arguments.

        This function exists in `feature-extraction/text.py`
        
        Parameters
        ----------
        list_args : str(s), optional
            Specific columns to apply this technique to.

        list_of_cols : list, optional
            A list of specific columns to apply this technique to., by default []

        keep_col : bool, optional
            True if you want to keep the column(s) or False if you want to drop the column(s)

        encoding: str, default=’utf-8’
            If bytes or files are given to analyze, this encoding is used to decode.

        decode_error: {‘strict’, ‘ignore’, ‘replace’} (default=’strict’)
            Instruction on what to do if a byte sequence is given to analyze that contains characters not of the given encoding.
            By default, it is ‘strict’, meaning that a UnicodeDecodeError will be raised.
            Other values are ‘ignore’ and ‘replace’.

        strip_accents: {‘ascii’, ‘unicode’, None} (default=None)
            Remove accents and perform other character normalization during the preprocessing step.\
            ‘ascii’ is a fast method that only works on characters that have an direct ASCII mapping.
            ‘unicode’ is a slightly slower method that works on any characters.
            None (default) does nothing.

            Both ‘ascii’ and ‘unicode’ use NFKD normalization from unicodedata.normalize.

        lowercase: bool (default=True)
            Convert all characters to lowercase before tokenizing.

        preprocessor: callable or None (default=None)
            Override the preprocessing (string transformation) stage while preserving the tokenizing and n-grams generation steps.
            Only applies if analyzer is not callable.

        tokenizer: callable or None (default=None)
            Override the string tokenization step while preserving the preprocessing and n-grams generation steps.
            Only applies if analyzer == 'word'.

        analyzer: str, {‘word’, ‘char’, ‘char_wb’} or callable
            Whether the feature should be made of word or character n-grams
            Option ‘char_wb’ creates character n-grams only from text inside word boundaries;
            n-grams at the edges of words are padded with space.

            If a callable is passed it is used to extract the sequence of features out of the raw, unprocessed input.

        stop_words: str {‘english’}, list, or None (default=None)
            If a string, it is passed to _check_stop_list and the appropriate stop list is returned.
            ‘english’ is currently the only supported string value.
            There are several known issues with ‘english’ and you should consider an alternative (see Using stop words).

            If a list, that list is assumed to contain stop words, all of which will be removed from the resulting tokens. Only applies if analyzer == 'word'.

            If None, no stop words will be used. max_df can be set to a value in the range [0.7, 1.0) to automatically detect and filter stop words based on intra corpus document frequency of terms.

        token_pattern: str
            Regular expression denoting what constitutes a “token”, only used if analyzer == 'word'.
            The default regexp selects tokens of 2 or more alphanumeric characters (punctuation is completely ignored and always treated as a token separator).
        
        ngram_range: tuple (min_n, max_n), default=(1, 1)
            The lower and upper boundary of the range of n-values for different n-grams to be extracted.
            All values of n such that min_n <= n <= max_n will be used.
            For example an ngram_range of (1, 1) means only unigrams, (1, 2) means unigrams and bigrams, and (2, 2) means only bigrams.
            Only applies if analyzer is not callable.
        
        max_df: float in range [0.0, 1.0] or int (default=1.0)
            When building the vocabulary ignore terms that have a document frequency strictly higher than the given threshold (corpus-specific stop words).
            If float, the parameter represents a proportion of documents, integer absolute counts.
            This parameter is ignored if vocabulary is not None.

        min_df: float in range [0.0, 1.0] or int (default=1)
            When building the vocabulary ignore terms that have a document frequency strictly lower than the given threshold.
            This value is also called cut-off in the literature.
            If float, the parameter represents a proportion of documents, integer absolute counts. This parameter is ignored if vocabulary is not None.

        max_features: int or None (default=None)
            If not None, build a vocabulary that only consider the top max_features ordered by term frequency across the corpus.

            This parameter is ignored if vocabulary is not None.

        vocabulary: Mapping or iterable, optional (default=None)
            Either a Mapping (e.g., a dict) where keys are terms and values are indices in the feature matrix, or an iterable over terms.
            If not given, a vocabulary is determined from the input documents.

        binary: bool (default=False)
            If True, all non-zero term counts are set to 1.
            This does not mean outputs will have only 0/1 values, only that the tf term in tf-idf is binary. (Set idf and normalization to False to get 0/1 outputs).

        dtype: type, optional (default=float64)
            Type of the matrix returned by fit_transform() or transform().

        norm: ‘l1’, ‘l2’ or None, optional (default=’l2’)
            Each output row will have unit norm, either: * ‘l2’: Sum of squares of vector elements is 1.
            The cosine similarity between two vectors is their dot product when l2 norm has been applied. * ‘l1’: Sum of absolute values of vector elements is 1.

        use_idf: bool (default=True)
            Enable inverse-document-frequency reweighting.

        smooth_idf: bool (default=True)
            Smooth idf weights by adding one to document frequencies, as if an extra document was seen containing every term in the collection exactly once.
            Prevents zero divisions.

        sublinear_tf: bool (default=False)
            Apply sublinear tf scaling, i.e. replace tf with 1 + log(tf).
        
        Returns
        -------
        Data:
            Returns a deep copy of the Data object.

        Examples
        --------
        >>> data.tfidf('col1', 'col2', 'col3')
        >>> data.tfidf('col1', 'col2', 'col3', lowercase=False, smoothidf=False)
        """

        report_info = technique_reason_repo["feature"]["text"]["tfidf"]

        ## If a list of columns is provided use the list, otherwise use arguemnts.
        list_of_cols = _input_columns(list_args, list_of_cols)

        self.x_train, self.x_test = text.feature_tfidf(
            x_train=self.x_train,
            x_test=self.x_test,
            list_of_cols=list_of_cols,
            keep_col=keep_col,
            **tfidf_kwargs,
        )

        if self.report is not None:
            self.report.report_technique(report_info, [])

        return self.copy()

    def bag_of_words(self, *list_args, list_of_cols=[], keep_col=True, **bow_kwargs):
        """
        Creates a matrix of how many times a word appears in a document.

        The premise is that the more times a word appears the more the word represents that document.

        For more information see: https://scikit-learn.org/stable/modules/generated/sklearn.feature_extraction.text.CountVectorizer.html

        If a list of columns is provided use the list, otherwise use arguments.

        This function exists in `feature-extraction/text.py`
        
        Parameters
        ----------
        list_args : str(s), optional
            Specific columns to apply this technique to.

        list_of_cols : list, optional
            A list of specific columns to apply this technique to., by default []

        keep_col : bool, optional
            True if you want to keep the column(s) or False if you want to drop the column(s)

        encoding: str, default=’utf-8’
            If bytes or files are given to analyze, this encoding is used to decode.

        decode_error: {‘strict’, ‘ignore’, ‘replace’} (default=’strict’)
            Instruction on what to do if a byte sequence is given to analyze that contains characters not of the given encoding.
            By default, it is ‘strict’, meaning that a UnicodeDecodeError will be raised.
            Other values are ‘ignore’ and ‘replace’.

        strip_accents: {‘ascii’, ‘unicode’, None} (default=None)
            Remove accents and perform other character normalization during the preprocessing step.\
            ‘ascii’ is a fast method that only works on characters that have an direct ASCII mapping.
            ‘unicode’ is a slightly slower method that works on any characters.
            None (default) does nothing.

            Both ‘ascii’ and ‘unicode’ use NFKD normalization from unicodedata.normalize.

        lowercase: bool (default=True)
            Convert all characters to lowercase before tokenizing.

        preprocessor: callable or None (default=None)
            Override the preprocessing (string transformation) stage while preserving the tokenizing and n-grams generation steps.
            Only applies if analyzer is not callable.

        tokenizer: callable or None (default=None)
            Override the string tokenization step while preserving the preprocessing and n-grams generation steps.
            Only applies if analyzer == 'word'.

        analyzer: str, {‘word’, ‘char’, ‘char_wb’} or callable
            Whether the feature should be made of word or character n-grams
            Option ‘char_wb’ creates character n-grams only from text inside word boundaries;
            n-grams at the edges of words are padded with space.

            If a callable is passed it is used to extract the sequence of features out of the raw, unprocessed input.

        stop_words: str {‘english’}, list, or None (default=None)
            If a string, it is passed to _check_stop_list and the appropriate stop list is returned.
            ‘english’ is currently the only supported string value.
            There are several known issues with ‘english’ and you should consider an alternative (see Using stop words).

            If a list, that list is assumed to contain stop words, all of which will be removed from the resulting tokens. Only applies if analyzer == 'word'.

            If None, no stop words will be used. max_df can be set to a value in the range [0.7, 1.0) to automatically detect and filter stop words based on intra corpus document frequency of terms.

        token_pattern: str
            Regular expression denoting what constitutes a “token”, only used if analyzer == 'word'.
            The default regexp selects tokens of 2 or more alphanumeric characters (punctuation is completely ignored and always treated as a token separator).
        
        ngram_range: tuple (min_n, max_n), default=(1, 1)
            The lower and upper boundary of the range of n-values for different n-grams to be extracted.
            All values of n such that min_n <= n <= max_n will be used.
            For example an ngram_range of (1, 1) means only unigrams, (1, 2) means unigrams and bigrams, and (2, 2) means only bigrams.
            Only applies if analyzer is not callable.
        
        max_df: float in range [0.0, 1.0] or int (default=1.0)
            When building the vocabulary ignore terms that have a document frequency strictly higher than the given threshold (corpus-specific stop words).
            If float, the parameter represents a proportion of documents, integer absolute counts.
            This parameter is ignored if vocabulary is not None.

        min_df: float in range [0.0, 1.0] or int (default=1)
            When building the vocabulary ignore terms that have a document frequency strictly lower than the given threshold.
            This value is also called cut-off in the literature.
            If float, the parameter represents a proportion of documents, integer absolute counts. This parameter is ignored if vocabulary is not None.

        max_features: int or None (default=None)
            If not None, build a vocabulary that only consider the top max_features ordered by term frequency across the corpus.

            This parameter is ignored if vocabulary is not None.

        vocabulary: Mapping or iterable, optional (default=None)
            Either a Mapping (e.g., a dict) where keys are terms and values are indices in the feature matrix, or an iterable over terms.
            If not given, a vocabulary is determined from the input documents.

        binary: bool (default=False)
            If True, all non-zero term counts are set to 1.
            This does not mean outputs will have only 0/1 values, only that the tf term in tf-idf is binary. (Set idf and normalization to False to get 0/1 outputs).
        
        Returns
        -------
        Data:
            Returns a deep copy of the Data object.

        Examples
        --------
        >>> data.bag_of_words('col1', 'col2', 'col3')
        >>> data.bag_of_words('col1', 'col2', 'col3', binary=True)
        """

        report_info = technique_reason_repo["feature"]["text"]["bow"]

        ## If a list of columns is provided use the list, otherwise use arguemnts.
        list_of_cols = _input_columns(list_args, list_of_cols)

        (self.x_train, self.x_test,) = text.feature_bag_of_words(
            x_train=self.x_train,
            x_test=self.x_test,
            list_of_cols=list_of_cols,
            keep_col=keep_col,
            **bow_kwargs,
        )

        if self.report is not None:
            self.report.report_technique(report_info, [])

        return self.copy()

    def text_hash(self, *list_args, list_of_cols=[], keep_col=True, **hash_kwargs):
        """
        Creates a matrix of how many times a word appears in a document. It can possibly normalized as token frequencies if norm=’l1’ or projected on the euclidean unit sphere if norm=’l2’.

        The premise is that the more times a word appears the more the word represents that document.

        This text vectorizer implementation uses the hashing trick to find the token string name to feature integer index mapping.

        This strategy has several advantages:

            It is very low memory scalable to large datasets as there is no need to store a vocabulary dictionary in memory
            It is fast to pickle and un-pickle as it holds no state besides the constructor parameters
            It can be used in a streaming (partial fit) or parallel pipeline as there is no state computed during fit.

        For more info please see: https://scikit-learn.org/stable/modules/generated/sklearn.feature_extraction.text.HashingVectorizer.html

        If a list of columns is provided use the list, otherwise use arguments.

        This function exists in `feature-extraction/text.py`
        
        Parameters
        ----------
        list_args : str(s), optional
            Specific columns to apply this technique to.

        list_of_cols : list, optional
            A list of specific columns to apply this technique to., by default []

        keep_col : bool, optional
            True if you want to keep the column(s) or False if you want to drop the column(s)

        n_features : integer, default=(2 ** 20)
            The number of features (columns) in the output matrices.
            Small numbers of features are likely to cause hash collisions, but large numbers will cause larger coefficient dimensions in linear learners.

        hash_kwargs : dict, optional
            Parameters you would pass into Bag of Words constructor, by default {}
        
        Returns
        -------
        Data:
            Returns a deep copy of the Data object.

        Examples
        --------
        >>> data.text_hash('col1', 'col2', 'col3')
        >>> data.text_hash('col1', 'col2', 'col3', n_features=50)
        """

        report_info = technique_reason_repo["feature"]["text"]["hash"]

        ## If a list of columns is provided use the list, otherwise use arguemnts.
        list_of_cols = _input_columns(list_args, list_of_cols)

        (self.x_train, self.x_test,) = text.feature_hash_vectorizer(
            x_train=self.x_train,
            x_test=self.x_test,
            list_of_cols=list_of_cols,
            keep_col=keep_col,
            **hash_kwargs,
        )

        if self.report is not None:
            self.report.report_technique(report_info, [])

        return self.copy()

    def postag_nltk(self, *list_args, list_of_cols=[], new_col_name="_postagged"):
        """
        Tag documents with their respective "Part of Speech" tag with the Textblob package which utilizes the NLTK NLP engine and Penn Treebank tag set.
        These tags classify a word as a noun, verb, adjective, etc. A full list and their meaning can be found here:
        https://www.ling.upenn.edu/courses/Fall_2003/ling001/penn_treebank_pos.html

        If a list of columns is provided use the list, otherwise use arguments.
        
        Parameters
        ----------
        list_args : str(s), optional
            Specific columns to apply this technique to.

        list_of_cols : list, optional
            A list of specific columns to apply this technique to., by default []

        new_col_name : str, optional
            New column name to be created when applying this technique, by default `COLUMN_postagged`

        Returns
        -------
        Data:
            Returns a deep copy of the Data object.

        Examples
        --------
        >>> data.postag_nltk('col1', 'col2', 'col3')
        """

        report_info = technique_reason_repo["feature"]["text"]["nltk_postag"]

        list_of_cols = _input_columns(list_args, list_of_cols)

        (self.x_train, self.x_test,) = text.nltk_feature_postag(
            x_train=self.x_train,
            x_test=self.x_test,
            list_of_cols=list_of_cols,
            new_col_name=new_col_name,
        )

        if self.report is not None:
            self.report.report_technique(report_info, [])

        return self.copy()

    def postag_spacy(self, *list_args, list_of_cols=[], new_col_name="_postagged"):
        """
        Tag documents with their respective "Part of Speech" tag with the Spacy NLP engine and the Universal Dependencies scheme.
        These tags classify a word as a noun, verb, adjective, etc. A full list and their meaning can be found here:
        https://spacy.io/api/annotation#pos-tagging 

        If a list of columns is provided use the list, otherwise use arguments.
        
        Parameters
        ----------
        list_args : str(s), optional
            Specific columns to apply this technique to.

        list_of_cols : list, optional
            A list of specific columns to apply this technique to., by default []

        new_col_name : str, optional
            New column name to be created when applying this technique, by default `COLUMN_postagged`

        Returns
        -------
        Data:
            Returns a deep copy of the Data object.

        Examples
        --------
        >>> data.postag_spacy('col1', 'col2', 'col3')
        """

        report_info = technique_reason_repo["feature"]["text"]["spacy_postag"]

        list_of_cols = _input_columns(list_args, list_of_cols)

        (self.x_train, self.x_test,) = text.spacy_feature_postag(
            x_train=self.x_train,
            x_test=self.x_test,
            list_of_cols=list_of_cols,
            new_col_name=new_col_name,
            method="s",
        )

        if self.report is not None:
            self.report.report_technique(report_info, list_of_cols)

        return self.copy()

    def postag_spacy_detailed(
        self, *list_args, list_of_cols=[], new_col_name="_postagged"
    ):
        """
        Tag documents with their respective "Part of Speech" tag with the Spacy NLP engine and the PennState PoS tags.
        These tags classify a word as a noun, verb, adjective, etc. A full list and their meaning can be found here:
        https://spacy.io/api/annotation#pos-tagging 

        If a list of columns is provided use the list, otherwise use arguments.
        
        Parameters
        ----------
        list_args : str(s), optional
            Specific columns to apply this technique to.

        list_of_cols : list, optional
            A list of specific columns to apply this technique to., by default []

        new_col_name : str, optional
            New column name to be created when applying this technique, by default `COLUMN_postagged`

        Returns
        -------
        Data:
            Returns a deep copy of the Data object.

        Examples
        --------
        >>> data.postag_spacy_detailed('col1', 'col2', 'col3')
        """

        report_info = technique_reason_repo["feature"]["text"]["spacy_postag"]

        list_of_cols = _input_columns(list_args, list_of_cols)

        (self.x_train, self.x_test,) = text.spacy_feature_postag(
            x_train=self.x_train,
            x_test=self.x_test,
            list_of_cols=list_of_cols,
            new_col_name=new_col_name,
            method="d",
        )

        if self.report is not None:
            self.report.report_technique(report_info, list_of_cols)

        return self.copy()

    def nounphrases_nltk(self, *list_args, list_of_cols=[], new_col_name="_phrases"):
        """
        Extract noun phrases from text using the Textblob packages which uses the NLTK NLP engine.

        If a list of columns is provided use the list, otherwise use arguments.
        
        Parameters
        ----------
        list_args : str(s), optional
            Specific columns to apply this technique to.

        list_of_cols : list, optional
            A list of specific columns to apply this technique to., by default []

        new_col_name : str, optional
            New column name to be created when applying this technique, by default `COLUMN_phrases`

        Returns
        -------
        Data:
            Returns a deep copy of the Data object.

        Examples
        --------
        >>> data.nounphrases_nltk('col1', 'col2', 'col3')
        """

        report_info = technique_reason_repo["feature"]["text"]["nltk_np"]

        list_of_cols = _input_columns(list_args, list_of_cols)

        (self.x_train, self.x_test,) = text.nltk_feature_noun_phrases(
            x_train=self.x_train,
            x_test=self.x_test,
            list_of_cols=list_of_cols,
            new_col_name=new_col_name,
        )

        if self.report is not None:
            self.report.report_technique(report_info, list_of_cols)

        return self.copy()

    def nounphrases_spacy(self, *list_args, list_of_cols=[], new_col_name="_phrases"):
        """
        Extract noun phrases from text using the Textblob packages which uses the NLTK NLP engine.

        If a list of columns is provided use the list, otherwise use arguments.
        
        Parameters
        ----------
        list_args : str(s), optional
            Specific columns to apply this technique to.

        list_of_cols : list, optional
            A list of specific columns to apply this technique to., by default []

        new_col_name : str, optional
            New column name to be created when applying this technique, by default `COLUMN_phrases`

        Returns
        -------
        Data:
            Returns a deep copy of the Data object.

        Examples
        --------
        >>> data.nounphrases_spacy('col1', 'col2', 'col3')
        """

        report_info = technique_reason_repo["feature"]["text"]["spacy_np"]

        list_of_cols = _input_columns(list_args, list_of_cols)

        (self.x_train, self.x_test,) = text.spacy_feature_noun_phrases(
            x_train=self.x_train,
            x_test=self.x_test,
            list_of_cols=list_of_cols,
            new_col_name=new_col_name,
        )

        if self.report is not None:
            self.report.report_technique(report_info, list_of_cols)

        return self.copy()

    def polynomial_features(self, *list_args, list_of_cols=[], **poly_kwargs):
        """
        Generate polynomial and interaction features.

        Generate a new feature matrix consisting of all polynomial combinations of the features with degree less than or equal to the specified degree.
        
        For example, if an input sample is two dimensional and of the form [a, b], the degree-2 polynomial features are [1, a, b, a^2, ab, b^2].
        
        Parameters
        ----------
        list_args : str(s), optional
            Specific columns to apply this technique to.

        list_of_cols : list, optional
            A list of specific columns to apply this technique to., by default []

        degree : int
            Degree of the polynomial features, by default 2

        interaction_only : boolean, 
            If true, only interaction features are produced: features that are products of at most degree distinct input features (so not x[1] ** 2, x[0] * x[2] ** 3, etc.).
            by default = False
        
        Returns
        -------
        Data:
            Returns a deep copy of the Data object.

        Examples
        --------
        >>> data.polynomial_features('col1', 'col2', 'col3')
        """

        report_info = technique_reason_repo["feature"]["numeric"]["poly"]

        ## If a list of columns is provided use the list, otherwise use arguemnts.
        list_of_cols = _input_columns(list_args, list_of_cols)

        (self.x_train, self.x_test,) = num.polynomial_features(
            x_train=self.x_train,
            x_test=self.x_test,
            list_of_cols=list_of_cols,
            **poly_kwargs,
        )

        if self.report is not None:
            self.report.report_technique(report_info, list_of_cols)

        return self.copy()

    def apply(self, func, output_col: str, description=""):
        """
        Calls pandas apply function. Will apply the function to your dataset, or
        both your training and testing dataset.
        
        Parameters
        ----------
        func : Function pointer
            Function describing the transformation for the new column

        output_col : str
            New column name
            
        description : str, optional
            Description of the new column to be logged into the report, by default ''
        
        Returns
        -------
        Feature:
            Returns a deep copy of the Feature object.

        Examples
        --------
        >>>     col1  col2  col3 
            0     1     0     1       
            1     0     2     0       
            2     1     0     1
        >>> data.apply(lambda x: x['col1'] > 0, 'col4')
        >>>     col1  col2  col3  col4 
            0     1     0     1     1       
            1     0     2     0     0  
            2     1     0     1     1
        """

        self.x_train, self.x_test = util.apply(
            x_train=self.x_train, func=func, output_col=output_col, x_test=self.x_test,
        )

        if self.report is not None:
            self.report.log(f"Added feature {output_col}. {description}")

        return self.copy()

    def encode_labels(self, *list_args, list_of_cols=[]):
        """
        Encode categorical values with value between 0 and n_classes-1.

        Running this function will automatically set the corresponding mapping for the target variable mapping number to the original value.

        Note that this will not work if your test data will have labels that your train data does not.        

        Parameters
        ----------
        list_args : str(s), optional
            Specific columns to apply this technique to.

        list_of_cols : list, optional
            A list of specific columns to apply this technique to., by default []
        
        Returns
        -------
        Data:
            Returns a deep copy of the Data object.

        Examples
        --------
        >>> data.encode_labels('col1', 'col2', 'col3')
        """

        report_info = technique_reason_repo["preprocess"]["categorical"]["label_encode"]

        list_of_cols = _input_columns(list_args, list_of_cols)

        self.x_train, self.x_test, _ = label_encoder(
            x_train=self.x_train, x_test=self.x_test, list_of_cols=list_of_cols,
        )

        if self.report is not None:
            self.report.report_technique(report_info, list_of_cols)

        return self.copy()

    def pca(self, n_components=10, **pca_kwargs):
        """
        Reduces the dimensionality of the data using Principal Component Analysis. 
        
        Use PCA when the data is dense.

        This can be used to reduce complexity as well as speed up computation.

        For more info please see: https://scikit-learn.org/stable/modules/generated/sklearn.decomposition.PCA.html 

        This function exists in `feature-extraction/util.py`
        
        Parameters
        ----------
        n_components : int, float, None or string, by default 10
            Number of components to keep.
            
            If n_components is not set all components are kept.
            If n_components == 'mle' and svd_solver == 'full', Minka’s MLE is used to guess the dimension.
            Use of n_components == 'mle' will interpret svd_solver == 'auto' as svd_solver == 'full'
            If 0 < n_components < 1 and svd_solver == 'full', select the number of components such that the amount of variance that needs to be explained is greater than the percentage specified by n_components
            If svd_solver == 'arpack', the number of components must be strictly less than the minimum of n_features and n_samples

        whiten : bool, optional (default False)
            When True (False by default) the components_ vectors are multiplied by the square root of n_samples and then divided by the singular values to ensure uncorrelated outputs with unit component-wise variances.
            Whitening will remove some information from the transformed signal (the relative variance scales of the components) but can sometime improve the predictive accuracy of the downstream estimators by making their data respect some hard-wired assumptions.

        svd_solver : string {‘auto’, ‘full’, ‘arpack’, ‘randomized’}
            auto :
                the solver is selected by a default policy based on X.shape and n_components: if the input data is larger than 500x500 and the number of components to extract is lower than 80% of the smallest dimension of the data, then the more efficient ‘randomized’ method is enabled. Otherwise the exact full SVD is computed and optionally truncated afterwards.
            full :
                run exact full SVD calling the standard LAPACK solver via scipy.linalg.svd and select the components by postprocessing
            arpack :
                run SVD truncated to n_components calling ARPACK solver via scipy.sparse.linalg.svds. It requires strictly 0 < n_components < min(X.shape)
            randomized :
                run randomized SVD by the method of Halko et al.

        tol : float >= 0, optional (default .0)
            Tolerance for singular values computed by svd_solver == ‘arpack’.

        iterated_power : int >= 0, or ‘auto’, (default ‘auto’)
            Number of iterations for the power method computed by svd_solver == ‘randomized’.

        random_state : int, RandomState instance or None, optional (default None)
            If int, random_state is the seed used by the random number generator;
            If RandomState instance, random_state is the random number generator;
            If None, the random number generator is the RandomState instance used by np.random.
            Used when svd_solver == ‘arpack’ or ‘randomized’.
        
        Returns
        -------
        Data:
            Returns a deep copy of the Data object.

        Examples
        --------
        >>> data.pca(n_components=2)
        """

        report_info = technique_reason_repo["feature"]["general"]["pca"]

        self._run_sklearn_dim_reduction("pca", n_components=n_components, **pca_kwargs)

        if self.report is not None:
            self.report.report_technique(report_info, [])

        return self.copy()

    def truncated_svd(self, n_components=50, **svd_kwargs):
        """
        Reduces the dimensionality of the data using Truncated SVD.

        In particular, truncated SVD works on term count/tf-idf matrices.
        In that context, it is known as latent semantic analysis (LSA).
        
        Use Truncated SVD when the data is sparse.

        This can be used to reduce complexity as well as speed up computation.

        For more info please see: https://scikit-learn.org/stable/modules/generated/sklearn.decomposition.TruncatedSVD.html

        This function exists in `feature-extraction/util.py`
        
        Parameters
        ---------- 
        n_components: int, default = 2
            Desired dimensionality of output data. Must be strictly less than the number of features.
            The default value is useful for visualisation. For LSA, a value of 100 is recommended.

        algorithm: string, default = “randomized”
            SVD solver to use. Either “arpack” for the ARPACK wrapper in SciPy (scipy.sparse.linalg.svds), or “randomized” for the randomized algorithm due to Halko (2009).

        n_iter: int, optional (default 5)
            Number of iterations for randomized SVD solver. Not used by ARPACK.
            The default is larger than the default in ~sklearn.utils.extmath.randomized_svd to handle sparse matrices that may have large slowly decaying spectrum.

        tol: float, optional
            Tolerance for ARPACK. 0 means machine precision. Ignored by randomized SVD solver.
        
        Returns
        -------
        Data:
            Returns a deep copy of the Data object.

        Examples
        --------
        >>> data.truncated_svd(n_components=2)
        """

        report_info = technique_reason_repo["feature"]["general"]["tsvd"]

        self._run_sklearn_dim_reduction("tsvd", n_components=n_components, **svd_kwargs)

        if self.report is not None:
            self.report.report_technique(report_info, [])

        return self.copy()

    def drop_correlated_features(self, threshold=0.95):
        """
        Drop features that have a correlation coefficient greater than the specified threshold with other features.
        
        Parameters
        ----------
        threshold : float, optional
            Correlation coefficient threshold, by default 0.95

        Returns
        -------
        Data:
            Returns a deep copy of the Data object.

        Examples
        --------
        >>> data.drop_correlated_features(threshold=0.9)
        """

        report_info = technique_reason_repo["clean"]["numeric"]["corr"]
        orig_cols = set(self.x_train.columns)

        (self.x_train, self.x_test,) = num.drop_correlated_features(
            x_train=self.x_train, x_test=self.x_test, threshold=threshold,
        )

        if self.report is not None:
            self.report.report_technique(
                report_info, list(orig_cols.difference(self.x_train.columns))
            )

        return self.copy()

    def chi2_feature_selection(self, k: int, verbose=False):
        """
        Uses Chi2 to choose the best K features.

        The Chi2 null hypothesis is that 2 variables are independent.

        Chi-square test feature selection “weeds out” the features that are the most likely to be independent of class and therefore irrelevant for classification.
        
        Parameters
        ----------
        k : int or "all"
            Number of features to keep.

        verbose : bool
            True to print p-values for each feature, by default False
        
        Returns
        -------
        Data:
            Returns a deep copy of the Data object.

        Examples
        --------
        >>> data.chi2_feature_selection(k=10)
        """

        report_info = technique_reason_repo["feature"]["numeric"]["chi2"]

        y_train = self.y_train
        y_test = self.y_test
        x_train = self.x_train.drop(self.target_field, axis=1)
        x_test = (
            None if self.x_test is None else self.x_test.drop(self.target_field, axis=1)
        )

        (self.x_train, self.x_test,) = num.kbest_chi2(
            x_train, self.y_train, x_test, k, verbose=verbose
        )

        # Re add the target field back to the dataset.
        self.x_train[self.target_field] = y_train
        if self.x_test is not None:
            self.x_test[self.target_field] = y_test

        if self.report is not None:
            self.report.report_technique(report_info)

        return self.copy()

    def _run_sklearn_dim_reduction(self, algo: str, n_components, **kwargs):
        """
        Generalized helper function to run dimensionality reduction algorithms from sklearn.
        
        Parameters
        ----------
        algo : str {'pca', 'tsne', 'lle', 'tsvd'}
            Dim Reduction algorithm to run.
        """

        if self.target_field:
            train_target_data = self.x_train[self.target_field]
            test_target_data = (
                self.x_test[self.target_field] if self.x_test is not None else None
            )
            x_train = self.x_train.drop(self.target_field, axis=1)
            x_test = (
                self.x_test.drop(self.target_field, axis=1)
                if self.x_test is not None
                else None
            )
        else:
            x_train = self.x_train
            x_test = self.x_test

        self.x_train, self.x_test = util.sklearn_dim_reduction(
            x_train=x_train,
            x_test=x_test,
            algo=algo,
            n_components=n_components,
            **kwargs,
        )

        if self.target_field:
            self.x_train[self.target_field] = train_target_data
            self.x_test[self.target_field] = (
                test_target_data if test_target_data is not None else None
            )



"""
This file contains the following functions:

sklearn_dim_reduction
apply
"""

import pandas as pd
from sklearn.decomposition import PCA, TruncatedSVD


def sklearn_dim_reduction(
    x_train, x_test=None, algo=None, n_components=50, **dim_reduce_kwargs
):
    """
    Performs Principal Component Analysis on a dataset.
    
    Parameters
    ----------
    x_train : DataFrame
        Dataset

    n_components : Any
        Number of components to reduce to - algo dependent.

    x_test : DataFrame
        Testing dataset, by default None
    
    Returns
    -------
    Dataframe, *Dataframe
        Transformed dataframe with the new column

    Returns 2 Dataframes if x_test is provided. 
    """

    algorithms = {
        "pca": PCA(n_components=n_components, **dim_reduce_kwargs),
        "tsvd": TruncatedSVD(n_components=n_components, **dim_reduce_kwargs),
    }

    reducer = algorithms[algo]

    x_train = pd.DataFrame(reducer.fit_transform(x_train))
    x_train.columns = map(str, x_train.columns)

    if x_test is not None:
        x_test = pd.DataFrame(reducer.transform(x_test))
        x_test.columns = map(str, x_test.columns)

    return x_train, x_test


def apply(x_train, func, output_col: str, x_test=None):
    """
    Wrapper for pandas apply function to be used in this library. Applies `func` to the entire data
    or just the trianing and testing data

    Parameters
    ----------
    x_train : DataFrame
        Dataset

    func : Function pointer
        Function describing the transformation for the new column

    output_col : str
        New column name
        
    x_test : DataFrame
        Testing dataset, by default None
    
    Returns
    -------
    Dataframe, *Dataframe
        Transformed dataframe with the new column

    Returns 2 Dataframes if x_test is provided. 
    """

    import swifter

    x_train.loc[:, output_col] = x_train.swifter.progress_bar().apply(func, axis=1)

    if x_test is not None:
        x_test.loc[:, output_col] = x_test.swifter.progress_bar().apply(func, axis=1)

    return x_train, x_test
